from django.core.exceptions import ValidationError

from ..base import TestCase
from ..testapp.models import (
    Address,
    Business,
    Owner,
    Part,
    Vehicle,
    VehicleType,
    db,
)


class TestCleanMixin(TestCase):
    maxDiff = None

    def test_full_clean_recursive(self):
        owner = Part(vehicles=[Vehicle(paint="pink", owner=Owner(first_name="invalid"))])

        self.assertRaisesWithMessageDict(
            owner,
            {"vehicles": [{"paint": ["Can't have a pink car"], "owner": {"first_name": ["Invalid first name"]}}]},
            recursive=True,
        )

    def test_full_clean_recursive_unloaded(self):
        owner = Part(vehicles=[Vehicle(type=VehicleType.car, paint="blue", owner=Owner(first_name="valid"))])
        db.add(owner)
        db.flush()

        # make owner invalid but pop from instance state
        # so full_clean() never validates owner
        owner.vehicles[0].owner.first_name = "invalid"
        owner.vehicles[0].__dict__.pop("owner")

        self.assertIsNone(owner.full_clean(recursive=True))

    def test_full_clean_recursive_empty(self):
        owner = Part(vehicles=[Vehicle(paint="pink", owner=None), None])

        self.assertRaisesWithMessageDict(owner, {"vehicles": [{"paint": ["Can't have a pink car"]}]}, recursive=True)

    def test_full_clean_exclude_specific_fields(self):
        owner = Part(vehicles=[Vehicle(paint="pink", owner=Owner(first_name="invalid"))])

        self.assertIsNone(owner.full_clean(recursive=True, exclude=["vehicles"]))
        self.assertIsNone(owner.full_clean(recursive=True, exclude={"vehicles": []}))
        self.assertIsNone(owner.full_clean(recursive=True, exclude={"vehicles": {}}))
        self.assertIsNone(owner.full_clean(recursive=True, exclude={"vehicles": ["paint", "owner"]}))
        self.assertIsNone(owner.full_clean(recursive=True, exclude={"vehicles": {"paint": None, "owner": {}}}))

    def test_full_clean_exclude_recursive_specific_fields(self):
        owner = Part(vehicles=[Vehicle(paint="pink", owner=Owner(first_name="invalid"))])

        self.assertRaisesWithMessageDict(
            owner,
            {"vehicles": [{"owner": {"first_name": ["Invalid first name"]}}]},
            recursive=True,
            exclude={"vehicles": ["paint"]},
        )
        self.assertRaisesWithMessageDict(
            owner,
            {"vehicles": [{"owner": {"first_name": ["Invalid first name"]}}]},
            recursive=True,
            exclude={"vehicles": {"paint": None}},
        )
        self.assertRaisesWithMessageDict(
            owner,
            {"vehicles": [{"paint": ["Can't have a pink car"]}]},
            recursive=True,
            exclude={"vehicles": {"owner": ["first_name"]}},
        )
        self.assertRaisesWithMessageDict(
            owner,
            {"vehicles": [{"paint": ["Can't have a pink car"]}]},
            recursive=True,
            exclude={"vehicles": {"owner": {"first_name": None}}},
        )

    def test_composite_full_clean(self):
        business = Business(location=None, other_location=Address(zip="01234", state="ny"))

        self.assertRaisesWithMessageDict(
            business,
            {
                "location": ["Primary location is required when other location is provided."],
                "other_location": {
                    "__all__": ["All state, street, zip are required."],
                    "state": ["Select a valid choice. ny is not one of the " "available choices."],
                    "zip": ["Zip cannot start with 0."],
                },
            },
            recursive=True,
        )
        self.assertRaisesWithMessageDict(
            business,
            {
                "location": ["Primary location is required when other location is provided."],
                "other_location": {
                    "__all__": ["All state, street, zip are required."],
                    "state": ["Select a valid choice. ny is not one of the available choices."],
                },
            },
            recursive=True,
            exclude={"other_location": {"zip": None}},
        )
        self.assertRaisesWithMessageDict(
            business,
            {
                "location": ["Primary location is required when other location is provided."],
                "other_location": {
                    "__all__": ["All state, street, zip are required."],
                    "state": ["Select a valid choice. ny is not one of the available choices."],
                },
            },
            recursive=True,
            exclude={"other_location": ["zip"]},
        )
        self.assertRaisesWithMessageDict(
            business,
            {"location": ["Primary location is required when other location is provided."]},
            recursive=True,
            exclude=["other_location"],
        )

    def assertRaisesWithMessageDict(self, model, expected, **kwargs):
        with self.assertRaises(ValidationError) as e:
            model.full_clean(**kwargs)

        self.assertEqual(e.exception.message_dict, expected)



import six

from django.conf import settings
from django.test import TestCase
from django_sorcery.db import databases
from django_sorcery.management.commands.sorcery_current import Command
from django_sorcery.management.commands.sorcery_downgrade import Downgrade
from django_sorcery.management.commands.sorcery_upgrade import Upgrade

from .base import MIGRATION_DIR, MigrationMixin


M1 = '''"""zero

Revision ID: 000000000000
Revises:
Create Date: 2018-07-24 01:28:40.893136

"""


# revision identifiers, used by Alembic.
revision = '000000000000'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
'''

M2 = '''"""one

Revision ID: 000000000001
Revises: 000000000000
Create Date: 2018-07-24 02:02:55.504526

"""


# revision identifiers, used by Alembic.
revision = '000000000001'
down_revision = '000000000000'
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
'''

db = databases.get("test")


class TestCurrent(MigrationMixin, TestCase):
    def setUp(self):
        super().setUp()

        for rev in ["000000000000", "000000000001"]:
            self.delete_migration("{}_.py".format(rev))

        self.write_migration(M1, "{}_.py".format("000000000000"))
        self.write_migration(M2, "{}_.py".format("000000000001"))

        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "tests_testapp", "--no-color"])

        Upgrade().run_from_argv(["./manage.py sorcery", "upgrade", "tests_testapp", "--no-color"])

    def tearDown(self):
        super().tearDown()

        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "tests_testapp", "--no-color"])

        for rev in ["000000000000", "000000000001"]:
            self.delete_migration("{}_.py".format(rev))

        Downgrade().run_from_argv(["./manage.py sorcery", "downgrade", "tests_testapp", "--no-color"])

    def test(self):
        out = six.StringIO()
        cmd = Command(stdout=out)
        cmd.run_from_argv(["./manage.py sorcery", "current", "--no-color"])

        out.seek(0)
        self.assertEqual(
            out.readlines(),
            [
                "Revision for tests_testapp on database test\n",
                "000000000001 (head)\n",
                "Revision for tests_otherapp on database test\n",
            ],
        )

    def test_verbose(self):
        out = six.StringIO()
        cmd = Command(stdout=out)
        cmd.run_from_argv(["./manage.py sorcery", "current", "-v", "2", "--no-color"])

        out.seek(0)
        self.assertEqual(
            out.readlines(),
            [
                "Revision for tests_testapp on database test\n",
                f"Current revision(s) for postgresql://postgres:***@{settings.DB_URL.host}/test\n",
                "Rev: 000000000001 (head)\n",
                "Parent: 000000000000\n",
                "Path: {}/000000000001_.py\n".format(MIGRATION_DIR),
                "\n",
                "    one\n",
                "    \n",
                "    Revision ID: 000000000001\n",
                "    Revises: 000000000000\n",
                "    Create Date: 2018-07-24 02:02:55.504526\n",
                "Revision for tests_otherapp on database test\n",
                f"Current revision(s) for postgresql://postgres:***@{settings.DB_URL.host}/test\n",
            ],
        )



"""Upgrade command."""
from functools import partial

import alembic
from django.core.management import CommandError

from ..alembic import AlembicCommand


class Upgrade(AlembicCommand):
    """Apply upgrade migration revisions."""

    help = "Apply migration revisions"

    def add_arguments(self, parser):
        parser.add_argument("app_label", nargs="?", help="App label of application to limit the output to.")
        parser.add_argument(
            "-r",
            "--revision",
            default="heads",
            help="Database state will be brought to the state after that "
            'migration. Use the name "heads" to apply all migrations.',
        )
        parser.add_argument(
            "-s",
            "--sql",
            default=False,
            action="store_true",
            help="Don't emit SQL to database - dump to standard output/file instead",
        )

    def handle(self, app_label=None, revision=None, sql=False, **kwargs):
        appconfigs = [self.lookup_app(app_label)] if app_label is not None else self.sorcery_apps.values()

        starting_rev = None
        if ":" in revision:
            starting_rev, revision = revision.split(":", 2)

        if revision != "heads" and app_label is None:
            raise CommandError("Revision requires an app_label to be provided")

        for appconfig in appconfigs:
            if not sql:
                self.stdout.write(
                    self.style.SUCCESS(
                        "Running migrations for {} on database {}".format(appconfig.name, appconfig.db.alias)
                    )
                )
            with alembic.context.EnvironmentContext(
                appconfig.config,
                appconfig.script,
                fn=partial(self.upgrade, appconfig=appconfig, revision=revision),
                as_sql=sql,
                starting_rev=starting_rev,
                destination_rev=revision,
            ) as context:
                self.run_env(context, appconfig)

    def upgrade(self, rev, context, appconfig, revision):
        """Executes alembic upgrade revisions to the given revision."""
        return appconfig.script._upgrade_revs(revision, rev)


Command = Upgrade



from django.test import TestCase
from django_sorcery import utils


class TestUtils(TestCase):
    def test_setdefaultattr(self):
        class Dummy:
            pass

        obj = Dummy()

        default = set()

        utils.setdefaultattr(obj, "test", default)

        self.assertTrue(hasattr(obj, "test"))
        self.assertEqual(obj.test, default)

        utils.setdefaultattr(obj, "test", set())

        self.assertTrue(hasattr(obj, "test"))
        self.assertEqual(obj.test, default)

    def test_make_args(self):
        value = utils.make_args("abc", kw="something")

        self.assertEqual(value, ("abc", {"kw": "something"}))

    def test_lower(self):
        self.assertEqual(utils.lower("HELLO"), "hello")
        self.assertEqual(utils.lower(5), 5)



"""History command."""

from django.core.management import CommandError

from ..alembic import AlembicCommand


class History(AlembicCommand):
    """Display alembic revisions."""

    help = "Display alembic revisions"

    def add_arguments(self, parser):
        parser.add_argument("app_label", nargs="?", help="App label of application to limit the output to.")
        parser.add_argument(
            "-r",
            "--rev_range",
            action="store",
            dest="rev_range",
            default=None,
            help="Specify a revision range; format is [start]:[end].",
        )

    def handle(self, app_label=None, rev_range=None, verbosity=0, **kwargs):
        verbose = bool(verbosity - 1)
        appconfigs = [self.lookup_app(app_label)] if app_label is not None else self.sorcery_apps.values()

        if rev_range is not None:
            if app_label is None:
                raise CommandError("Revision requires an app label")

            if ":" not in rev_range:
                raise CommandError("History range requires [start]:[end], [start]:, or :[end]")
            base, head = rev_range.strip().split(":")
        else:
            base = head = None

        self.print_history(appconfigs, verbose, base, head)

    def print_history(self, appconfigs, verbose, base, head):
        """Prints alembic revision history."""
        for appconfig in appconfigs:
            self.stdout.write(
                self.style.SUCCESS("Migrations for {} on database {}".format(appconfig.name, appconfig.db.alias))
            )
            for rev in appconfig.script.walk_revisions(base=base or "base", head=head or "heads"):
                self.stdout.write(
                    rev.cmd_format(verbose=verbose, include_branches=True, include_doc=True, include_parents=True)
                )


Command = History



from django_sorcery.db import databases


db = databases.get("minimal_backpop")


class Asset(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))

    order = db.ManyToOne("Order", back_populates="assets")


class OrderItem(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))

    orders = db.OneToMany("Order", back_populates="order_item")


class Customer(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))

    orders = db.OneToMany("Order", back_populates="applicant")
    coapp_orders = db.OneToMany("Order", back_populates="coapplicant")
    profile = db.relationship("Profile", back_populates="customer", uselist=False)


class Profile(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=20))

    customer = db.OneToOne("Customer", back_populates="profile")


class Contact(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))

    orders = db.ManyToMany("Order", back_populates="contacts", table_name="order_contacts")


class Order(db.Model):
    pk = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String(length=5))

    order_item = db.ManyToOne("OrderItem", back_populates="orders")

    applicant = db.ManyToOne("Customer", back_populates="orders")

    coapplicant = db.ManyToOne("Customer", back_populates="coapp_orders")

    assets = db.OneToMany(Asset, back_populates="order")

    contacts = db.ManyToMany("Contact", back_populates="orders", table_name="order_contacts")


db.configure_mappers()
db.drop_all()
db.create_all()



"""sqlalchemy transaction related things."""
import functools


class TransactionContext:
    """Transaction context manager for maintaining a transaction or
    savepoint."""

    def __init__(self, *dbs, **kwargs):
        self.dbs = dbs
        self.savepoint = kwargs.get("savepoint", True)
        self.transactions = None

    def __call__(self, func, *args, **kwargs):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapped

    def __enter__(self):
        self.transactions = [db.begin(subtransactions=True, nested=self.savepoint) for db in self.dbs]
        return self

    def __exit__(self, exception_type, value, tb=None):
        if exception_type is None:
            try:
                for transaction in self.transactions:
                    transaction.session.flush()
            except Exception as ex:
                exception_type = ex.__class__
                value = ex
                tb = getattr(ex, "__traceback__", None)
        for transaction in self.transactions:
            transaction.__exit__(exception_type, value, tb)
        self.transactions = None
        if value:
            raise value.with_traceback(tb)



"""Default alembic config things."""

from ...db import signals


@signals.alembic_config_created.connect
def setup_config(config):
    """The default alembic config created handler."""
    config.set_section_option("loggers", "keys", "root,sqlalchemy,alembic")
    config.set_section_option("handlers", "keys", "console")
    config.set_section_option("formatters", "keys", "generic")
    config.set_section_option("logger_root", "level", "WARN")
    config.set_section_option("logger_root", "handlers", "console")
    config.set_section_option("logger_sqlalchemy", "level", "WARN")
    config.set_section_option("logger_sqlalchemy", "qualname", "sqlalchemy.engine")
    config.set_section_option("logger_alembic", "level", "INFO")
    config.set_section_option("logger_alembic", "qualname", "alembic")
    config.set_section_option("handler_console", "class", "StreamHandler")
    config.set_section_option("handler_console", "formatter", "generic")



from django_sorcery.db import databases


default_db = databases.get("default")
other_db = databases.get("fromdbs")


class Foo(default_db.Model):
    id = default_db.Column(default_db.Integer(), autoincrement=True, primary_key=True)
    name = default_db.Column(default_db.String(length=10), nullable=False)


class Bar(other_db.Model):
    id = other_db.Column(other_db.Integer(), autoincrement=True, primary_key=True)
    name = other_db.Column(other_db.String(length=10), nullable=False)


default_db.create_all()
other_db.create_all()



"""Helper functions for creating Form classes from SQLAlchemy models."""
from collections import OrderedDict
from contextlib import suppress
from itertools import chain

from django.core.exceptions import (
    NON_FIELD_ERRORS,
    ImproperlyConfigured,
    ValidationError,
)
from django.forms import ALL_FIELDS
from django.forms.forms import (
    BaseForm as DjangoBaseForm,
    DeclarativeFieldsMetaclass,
)
from django.forms.models import (
    BaseModelForm as DjangoBaseModelForm,
    ModelFormOptions,
)
from django.forms.utils import ErrorList

from .db import meta


def _get_default_kwargs(
    info,
    session,
    fields=None,
    exclude=None,
    widgets=None,
    localized_fields=None,
    labels=None,
    help_texts=None,
    error_messages=None,
    field_classes=None,
):
    kwargs = {}
    if widgets and info.name in widgets:
        kwargs["widget"] = widgets[info.name]
    if labels and info.name in labels:
        kwargs["label"] = labels[info.name]
    if help_texts and info.name in help_texts:
        kwargs["help_text"] = help_texts[info.name]
    if error_messages and info.name in error_messages:
        kwargs["error_messages"] = error_messages[info.name]
    if field_classes and info.name in field_classes:
        kwargs["form_class"] = field_classes[info.name]

    if localized_fields == ALL_FIELDS:
        kwargs["localize"] = True
    if localized_fields and info.name in localized_fields:
        kwargs["localize"] = True

    if session is None:
        with suppress(AttributeError):
            session = info.related_model.query.session

    if isinstance(info, meta.relation_info):
        kwargs["session"] = session
    return kwargs


def fields_for_model(
    model,
    session,
    fields=None,
    exclude=None,
    widgets=None,
    formfield_callback=None,
    localized_fields=None,
    labels=None,
    help_texts=None,
    error_messages=None,
    field_classes=None,
    apply_limit_choices_to=True,
    **kwargs,
):
    """Returns a dictionary containing form fields for a given model."""

    field_list = []
    info = meta.model_info(model)

    for name, attr in chain(info.properties.items(), info.relationships.items()):
        if name.startswith("_"):
            continue

        if fields and name not in fields:
            continue

        if exclude and name in exclude:
            continue

        kwargs = _get_default_kwargs(
            attr,
            session,
            fields=fields,
            exclude=exclude,
            widgets=widgets,
            localized_fields=localized_fields,
            labels=labels,
            help_texts=help_texts,
            error_messages=error_messages,
            field_classes=field_classes,
        )
        if formfield_callback is None:
            formfield = attr.formfield(**kwargs)
        elif not callable(formfield_callback):
            raise TypeError("formfield_callback must be a function or callable")
        else:
            formfield = formfield_callback(attr, **kwargs)

        if formfield is not None:
            if apply_limit_choices_to:
                apply_limit_choices_to_form_field(formfield)
            field_list.append((name, formfield))

    return OrderedDict(field_list)


def apply_limit_choices_to_form_field(formfield):
    """Apply limit_choices_to to the formfield's query if needed."""
    if hasattr(formfield, "queryset") and hasattr(formfield, "get_limit_choices_to"):
        limit_choices_to = formfield.get_limit_choices_to()
        if limit_choices_to is not None:
            formfield.queryset = formfield.queryset.filter(*limit_choices_to)


def model_to_dict(instance, fields=None, exclude=None):
    """Return a dict containing the data in ``instance`` suitable for passing
    as a Form's ``initial`` keyword argument.

    ``fields`` is an optional list of field names. If provided, return only the
    named.

    ``exclude`` is an optional list of field names. If provided, exclude the
    named from the returned dict, even if they are listed in the ``fields``
    argument.
    """
    info = meta.model_info(type(instance))

    fields = set(
        fields or list(info.properties.keys()) + list(info.primary_keys.keys()) + list(info.relationships.keys())
    )
    exclude = set(exclude or [])
    data = {}
    for name in info.properties:
        if name.startswith("_"):
            continue

        if name not in fields:
            continue

        if name in exclude:
            continue

        data[name] = getattr(instance, name)

    for name, rel in info.relationships.items():
        related_info = meta.model_info(rel.related_model)

        if name.startswith("_"):
            continue

        if name not in fields:
            continue

        if name in exclude:
            continue

        if rel.uselist:
            for obj in getattr(instance, name):
                pks = related_info.primary_keys_from_instance(obj)
                if pks:
                    data.setdefault(name, []).append(pks)
        else:
            obj = getattr(instance, name)
            pks = related_info.primary_keys_from_instance(obj)
            if pks:
                data[name] = pks

    return data


class SQLAModelFormOptions(ModelFormOptions):
    """Model form options for sqlalchemy."""

    def __init__(self, options=None):
        super().__init__(options=options)
        self.session = getattr(options, "session", None)


class ModelFormMetaclass(DeclarativeFieldsMetaclass):
    """ModelForm metaclass for sqlalchemy models."""

    def __new__(mcs, name, bases, attrs):
        cls = super().__new__(mcs, name, bases, attrs)

        base_formfield_callback = next(
            (
                base.Meta.formfield_callback
                for base in bases
                if hasattr(base, "Meta") and hasattr(base.Meta, "formfield_callback")
            ),
            None,
        )
        formfield_callback = attrs.pop("formfield_callback", base_formfield_callback)

        if bases == (BaseModelForm,):
            return cls

        opts = cls._meta = SQLAModelFormOptions(getattr(cls, "Meta", None))

        for opt in ["fields", "exclude", "localized_fields"]:
            value = getattr(opts, opt)
            if isinstance(value, str) and value != ALL_FIELDS:
                raise TypeError(
                    "%(model)s.Meta.%(opt)s cannot be a string. Did you mean to type: ('%(value)s',)?"
                    % {"model": cls.__name__, "opt": opt, "value": value}
                )

        if opts.model:
            if opts.fields is None and opts.exclude is None:
                raise ImproperlyConfigured(
                    "Creating a ModelForm without either the 'fields' attribute or the 'exclude' "
                    f"attribute is prohibited; form {name} needs updating."
                )

            if opts.fields == ALL_FIELDS:
                opts.fields = None

            cls.base_fields = fields_for_model(
                opts.model,
                opts.session,
                error_messages=opts.error_messages,
                exclude=opts.exclude,
                field_classes=opts.field_classes,
                fields=opts.fields,
                help_texts=opts.help_texts,
                labels=opts.labels,
                localized_fields=opts.localized_fields,
                widgets=opts.widgets,
                formfield_callback=formfield_callback,
                apply_limit_choices_to=False,
            )
            cls.base_fields.update(cls.declared_fields)
        else:
            cls.base_fields = cls.declared_fields

        return cls


class BaseModelForm(DjangoBaseModelForm):
    """Base ModelForm for sqlalchemy models."""

    def __init__(
        self,
        data=None,
        files=None,
        auto_id="id_%s",
        prefix=None,
        initial=None,
        error_class=ErrorList,
        label_suffix=None,
        empty_permitted=False,
        instance=None,
        use_required_attribute=None,
        renderer=None,
        session=None,
    ):
        opts = self._meta
        opts.session = opts.session or session
        if opts.model is None:
            raise ValueError("ModelForm has no model class specified.")

        if opts.session is None:
            raise ValueError("ModelForm has no session specified.")

        self.instance = opts.model() if instance is None else instance
        object_data = self.model_to_dict()
        object_data.update(initial or {})
        self._validate_unique = False
        DjangoBaseForm.__init__(
            self,
            data=data,
            files=files,
            auto_id=auto_id,
            prefix=prefix,
            initial=object_data,
            error_class=error_class,
            label_suffix=label_suffix,
            empty_permitted=empty_permitted,
            use_required_attribute=use_required_attribute,
            renderer=renderer,
        )
        for field in self.fields.values():
            apply_limit_choices_to_form_field(field)

    def model_to_dict(self):
        """Returns a dict containing the data in ``instance`` suitable for
        passing as forms ``initial`` keyword argument."""
        opts = self._meta
        return model_to_dict(self.instance, opts.fields, opts.exclude)

    def _update_errors(self, errors):
        custom_errors = self._meta.error_messages or {}
        error_dict = getattr(errors, "error_dict", None) or {NON_FIELD_ERRORS: errors}

        for field, messages in error_dict.items():
            error_messages = {}
            if field == NON_FIELD_ERRORS and NON_FIELD_ERRORS in custom_errors:
                error_messages = custom_errors[NON_FIELD_ERRORS]
            elif field in self.fields:
                error_messages = self.fields[field].error_messages

            for message in messages:
                if isinstance(message, ValidationError) and message.code in error_messages:
                    message.message = error_messages[message.code]

        self.add_error(None, errors)

    def is_valid(self, rollback=True):
        """Return True if the form has no errors, or False otherwise.

        Will also rollback the session transaction.
        """
        is_valid = super().is_valid()

        if not is_valid and rollback:
            self._meta.session.rollback()

        return is_valid

    def save(self, flush=True, **kwargs):
        """Makes form's self.instance model persistent and flushes the
        session."""
        opts = self._meta

        if self.errors:
            raise ValueError(
                f"The {self.instance.__class__.__name__} could not be saved because the data didn't validate."
            )

        if self.instance not in opts.session:
            opts.session.add(self.instance)

        if flush:
            opts.session.flush()

        return self.instance

    def _post_clean(self):
        """Hook for performing additional cleaning after form cleaning is
        complete.

        Used for model validation in model forms.
        """
        try:
            self.instance = self.save_instance()
        except ValidationError as e:
            self._update_errors(e)

        try:
            getattr(self.instance, "full_clean", bool)()
        except ValidationError as e:
            self._update_errors(e)

    def save_instance(self, instance=None, cleaned_data=None):
        """Updates form's instance with cleaned data."""
        instance = instance or self.instance
        cleaned_data = cleaned_data or self.cleaned_data

        for name, field in self.fields.items():
            if name in cleaned_data:
                self.update_attribute(self.instance, name, field, cleaned_data[name])

        return instance

    def update_attribute(self, instance, name, field, value):
        """Provides hooks for updating form instance's attribute for a field
        with value."""
        field_setter = getattr(self, f"set_{name}", None)
        if field_setter:
            field_setter(instance, name, field, value)
        else:
            setattr(instance, name, value)


class ModelForm(BaseModelForm, metaclass=ModelFormMetaclass):
    """ModelForm base class for sqlalchemy models."""


def modelform_factory(model, form=ModelForm, formfield_callback=None, **kwargs):
    """Return a ModelForm class containing form fields for the given model."""
    defaults = [
        "fields",
        "exclude",
        "widgets",
        "localized_fields",
        "labels",
        "help_texts",
        "error_messages",
        "field_classes",
        "session",
    ]

    attrs = {"model": model}
    for key in defaults:
        value = kwargs.get(key)
        if value is not None:
            attrs[key] = value

    bases = (form.Meta,) if hasattr(form, "Meta") else ()
    meta_ = type("Meta", bases, attrs)
    if formfield_callback:
        meta_.formfield_callback = staticmethod(formfield_callback)

    class_name = f"{model.__name__}Form"

    if getattr(meta_, "fields", None) is None and getattr(meta_, "exclude", None) is None:
        raise ImproperlyConfigured(
            "Calling modelform_factory without defining 'fields' or 'exclude' explicitly is prohibited."
        )

    return type(form)(str(class_name), (form,), {"Meta": meta_, "formfield_callback": formfield_callback})



"""sqlalchemy session related things."""
from itertools import chain

from sqlalchemy import event, orm

from ..utils import setdefaultattr
from . import signals


def before_flush(session, flush_context, instances):
    signals.before_flush.send(session, flush_context=flush_context, instances=instances)


def after_flush(session, flush_context):
    signals.after_flush.send(session, flush_context=flush_context)


def before_commit(session):
    if session.transaction and (session.transaction._parent is None or not session.transaction.nested):
        signals.before_commit.send(session)
        signals.before_scoped_commit.send(session)


def after_commit(session):
    if session.transaction and (session.transaction._parent is None or not session.transaction.nested):
        signals.after_scoped_commit.send(session)
        signals.after_commit.send(session)
        setdefaultattr(session, "models_committed", set()).clear()
        setdefaultattr(session, "models_deleted", set()).clear()


def after_rollback(session):
    if session.transaction and (session.transaction._parent is None or session.transaction.nested):
        signals.after_scoped_rollback.send(session)
        signals.after_rollback.send(session)
        setdefaultattr(session, "models_committed", set()).clear()
        setdefaultattr(session, "models_deleted", set()).clear()


def record_models(session, flush_context=None, instances=None):
    setdefaultattr(session, "models_committed", set())
    setdefaultattr(session, "models_deleted", set())

    for instance in chain(session.new, session.dirty):
        session.models_committed.add(instance)

    for instance in session.deleted:
        session.models_deleted.add(instance)
        if instance in session.models_committed:
            session.models_committed.remove(instance)


class SignallingSession(orm.Session):
    """A custom sqlalchemy session implementation that provides signals."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        event.listen(self, "after_flush", record_models)

        event.listen(self, "before_flush", before_flush)
        event.listen(self, "after_flush", after_flush)
        event.listen(self, "before_commit", before_commit)
        event.listen(self, "after_commit", after_commit)
        event.listen(self, "after_rollback", after_rollback)

    def query(self, *args, **kwargs):
        """Override to try to use the model.query_class."""
        if len(args) == 1 and hasattr(args[0], "query_class") and args[0].query_class is not None:
            return args[0].query_class(*args, session=self, **kwargs)

        return super().query(*args, **kwargs)



from django.urls import reverse_lazy
from django_sorcery import forms, views, viewsets
from django_sorcery.routers import action

from .models import Owner, Vehicle, db


class OwnerListView(views.ListView):
    queryset = Owner.query
    paginate_by = 20


class OwnerListViewWithTemplate(views.ListView):
    template_name = "views/list.html"
    queryset = Owner.query


class OwnerListViewWithOrdering(views.ListView):
    queryset = Owner.query
    ordering = (Owner.id.desc(),)


class OwnerListViewNoContextName(views.ListView):
    queryset = Owner.query
    context_object_name = None


class OwnerListViewNoEmptyPaginateBy(views.ListView):
    queryset = Owner.query
    paginate_by = 20
    allow_empty = False


class DummyListViewNoEmptyPaginateBy(views.ListView):
    queryset = []
    paginate_by = 20
    allow_empty = False


class OwnerDetailView(views.DetailView):
    queryset = Owner.query


class OwnerDetailViewWithTemplate(views.DetailView):
    template_name = "views/edit.html"
    queryset = Owner.query


class OwnerDetailViewWithSlug(views.DetailView):
    queryset = Owner.query
    query_pkg_and_slug = True
    slug_field = Owner.last_name.key


class OwnerDetailViewContextName(views.DetailView):
    queryset = Owner.query
    context_object_name = "item"


class OwnerCreateView(views.CreateView):
    model = Owner
    fields = "__all__"
    session = db
    success_url = reverse_lazy("owners_list")


class OwnerUpdateView(views.UpdateView):
    model = Owner
    fields = "__all__"
    session = db
    success_url = reverse_lazy("owners_list")


class OwnerCreateViewWithFieldsAndForm(views.CreateView):
    model = Owner
    session = db
    fields = "__all__"
    form_class = forms.modelform_factory(Owner, fields="__all__", session=db)


class OwnerCreateViewWithForm(views.CreateView):
    model = Owner
    session = db
    form_class = forms.modelform_factory(Owner, fields="__all__", session=db)


class VehicleCreateView(views.CreateView):
    model = Vehicle
    session = db
    fields = "__all__"


class OwnerDeleteView(views.DeleteView):
    model = Owner
    session = db
    success_url = reverse_lazy("owners_list")


class OwnerViewSet(viewsets.ModelViewSet):
    model = Owner

    @action(detail=True)
    def custom(self, request, *args, **kwargs):
        return self.detail(request, *args, **kwargs)



"""Package with SQLAlchemy abstractions to interact with the database. All
tools implemented here assume `unit-of-work`_ usage pattern.

Connecting
----------

Connecting to the DB is as simple as using a url with :py:class:`.SQLAlchemy` abstraction::

    >>> from django_sorcery.db import SQLAlchemy
    >>> db = SQLAlchemy('sqlite://')

You can also use an alias to connect to a database as well. Just like Django's ``DATABASES``, connection settings for
aliases are stored in ``SQLALCHEMY_CONNECTIONS``. As such all databases are referred by their aliases just like in
Django ORM. For example ``"default"`` is an alias to the default connect::

    >>> db = SQLAlchemy('default')

ORM Goodies
-----------

:py:class:`.SQLAlchemy` is itself also a threadlocal manager that proxies calls to a thread local ``Session`` instance
and provides a couple of shortcuts when interacting with SQLAlchemy session.
To have them installed, you can subclass from the declarative base generated by the DB.
In addition :py:class:`.SQLAlchemy` also exposes most of SQLALchemy's elements
so a single import should suffice to define most tables::

    >>> class BarModel(db.Model):
    ...     id = db.Column(db.Integer(), primary_key=True)

    >>> class FooModel(db.Model):
    ...     id = db.Column(db.Integer(), primary_key=True)
    ...     id2 = db.Column(db.Integer(), primary_key=True)
    ...     name = db.Column(db.String(length=32))
    ...     bar_id = db.Column(db.Integer(), db.ForeignKey(BarModel.id, name='FK_foo_bar', use_alter=True))
    ...     bar = db.relationship(BarModel)

    >>> db.create_all()

Doing so will allow to use a couple of useful shortcuts:

:py:meth:`.Query.get`:

    Identity map in SQLAlchemy is amazing. Composite keys not so much. Put them together is even worse.
    :py:attr:`.SQLAlchemy` uses a special session which allows keyword arguments when using ``.get()``.
    It is much more explicit and is less error-prone as SQLAlchemy's default implementation expects only
    positional arguments where order of primary keys matters.
    For example::

        >>> db.query(FooModel).get(id=123, id2=456)

:py:meth:`.SQLAlchemy.queryproperty()`:

    Allows to create a property which will normalize to a query object of the model.
    It will use the correct session within the transaction so no need to pass session around.
    For example::

        >>> class MyView(object):
        ...     queryset = db.queryproperty(FooModel)

    You can even pass default filtering criteria if needed::

        >>> class MyView(object):
        ...     queryset = db.queryproperty(FooModel, to_be_deleted=False)

    In addition this pattern can be used to implement Django's ORM style model managers::

        >>> class UserModel(db.Model):
        ...     id = db.Column(db.Integer(), primary_key=True)
        ...     username = db.Column(db.String())
        ...     is_active = db.Column(db.Boolean())
        ...
        ...     objects = db.queryproperty()
        ...     active = db.queryproperty(is_active=True)

    That can be used directly::

        >>> UserModel.metadata.create_all(bind=db.engine)
        >>> db.add_all([
        ...     UserModel(id=1, username='foo', is_active=False),
        ...     UserModel(id=2, username='bar', is_active=True),
        ... ])
        >>> db.flush()

        >>> UserModel.objects.all()
        [UserModel(id=1, is_active=False, username='foo'), UserModel(id=2, is_active=True, username='bar')]
        >>> UserModel.active.all()
        [UserModel(id=2, is_active=True, username='bar')]

    This pattern is very useful when combined with Django style views::

        >>> class MyView(object):
        ...     queryset = UserModel.active

        >>> MyView().queryset.all()
        [UserModel(id=2, is_active=True, username='bar')]

    Additional filters/options can be applied as well::

        >>> class MyView(object):
        ...     queryset = UserModel.active.filter(UserModel.username == 'test')

        >>> MyView().queryset.all()
        []

Transactions
------------

If you need to explicitly use transactions, you can use :py:meth:`.SQLAlchemy.atomic`::

    >>> with db.atomic(savepoint=False):
    ...     _ = db.query(UserModel).filter_by(username='hello').update(dict(username='world'))

    >>> @db.atomic(savepoint=False)
    ... def do_something():
    ...     db.query(UserModel).filter_by(username='hello').update(username='world')

.. warning::

    You better know what you are doing if you need this.
    If you are not, sure, **JUST USE THE MIDDLEWARE** (see below).

Django
------

To complete the unit-of-work :py:class:`.SQLAlchemy` ships with capability
to generate Django middleware to commit at the end of the request
for the used session within the request::

    # settings.py
    >>> MIDDLEWARE = [
    ...     # should be last in response life-cycle
    ...     # or first in middleware list
    ...     'path.to.db.middleware',
    ...     # rest of middleware
    ... ]

.. _unit-of-work: https://martinfowler.com/eaaCatalog/unitOfWork.html
"""


from .sqlalchemy import SQLAlchemy  # noqa
from .utils import dbdict


databases = dbdict()



from django_sorcery.db import meta
from django_sorcery.forms import ALL_FIELDS
from django_sorcery.formsets import inlineformset_factory

from ..base import TestCase
from ..testapp.models import Owner, Vehicle, VehicleType, db


class TestInlineFormSet(TestCase):
    def setUp(self):
        super().setUp()
        self.owner = Owner(first_name="Test 1", last_name="Owner 1")
        db.add(self.owner)
        db.flush()

    def test_factory(self):
        info = meta.model_info(Owner)
        formset_class = inlineformset_factory(Owner, Vehicle, fields=ALL_FIELDS, session=db)

        self.assertEqual(formset_class.fk, info.relationships["vehicles"])

    def test_factory_with_relation(self):
        info = meta.model_info(Owner)
        formset_class = inlineformset_factory(relation=Owner.vehicles, fields=ALL_FIELDS, session=db)

        self.assertEqual(formset_class.fk, info.relationships["vehicles"])

    def test_factory_with_bad_relation(self):
        with self.assertRaises(ValueError):
            inlineformset_factory(relation=Vehicle.owner, fields=ALL_FIELDS, session=db)

    def test_factory_for_non_existant_relation(self):
        with self.assertRaises(ValueError):
            inlineformset_factory(Vehicle, Owner, fields=ALL_FIELDS, session=db)

    def test_initial_form_count(self):
        formset_class = inlineformset_factory(relation=Owner.vehicles, fields=("type",), session=db)
        formset = formset_class(save_as_new=True)
        self.assertEqual(formset.initial_form_count(), 0)

    def test_inline_form_create_render(self):
        formset_class = inlineformset_factory(relation=Owner.vehicles, fields=("type",), session=db)
        formset = formset_class()

        soup = self.get_soup(formset.as_p())
        expected_soup = self.get_soup(
            "".join(
                [
                    '<input type="hidden" name="vehicles-TOTAL_FORMS" value="3" id="id_vehicles-TOTAL_FORMS" />',
                    '<input type="hidden" name="vehicles-INITIAL_FORMS" value="0" id="id_vehicles-INITIAL_FORMS" />',
                    '<input type="hidden" name="vehicles-MIN_NUM_FORMS" value="0" id="id_vehicles-MIN_NUM_FORMS" />',
                    '<input type="hidden" name="vehicles-MAX_NUM_FORMS" value="1000" id="id_vehicles-MAX_NUM_FORMS" />',
                    "<p>",
                    '  <label for="id_vehicles-0-type">Type:</label>',
                    '  <select id="id_vehicles-0-type" name="vehicles-0-type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-0-DELETE">Delete:</label>',
                    '  <input id="id_vehicles-0-DELETE" name="vehicles-0-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-1-type">Type:</label>',
                    '  <select id="id_vehicles-1-type" name="vehicles-1-type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-1-DELETE">Delete:</label>',
                    '  <input id="id_vehicles-1-DELETE" name="vehicles-1-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-2-type">Type:</label>',
                    '  <select id="id_vehicles-2-type" name="vehicles-2-type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-2-DELETE">Delete:</label>',
                    '  <input id="id_vehicles-2-DELETE" name="vehicles-2-DELETE" type="checkbox" />',
                    "</p>",
                ]
            )
        )
        self.assertHTMLEqual(soup.prettify(), expected_soup.prettify())

    def test_inline_form_create(self):
        formset_class = inlineformset_factory(relation=Owner.vehicles, fields=("type",), session=db)
        data = {
            "vehicles-TOTAL_FORMS": "3",
            "vehicles-INITIAL_FORMS": "3",
            "vehicles-MIN_NUM_FORMS": "0",
            "vehicles-MAX_NUM_FORMS": "1000",
            "vehicles-0-type": "car",
            "vehicles-1-type": "bus",
            "vehicles-2-type": "car",
        }
        formset = formset_class(instance=self.owner, data=data)
        values = formset.save()

        self.assertEqual(values[0].owner, self.owner)
        self.assertEqual(values[0].type, VehicleType.car)

        self.assertEqual(values[1].owner, self.owner)
        self.assertEqual(values[1].type, VehicleType.bus)

        self.assertEqual(values[2].owner, self.owner)
        self.assertEqual(values[2].type, VehicleType.car)

    def test_inline_form_update_render(self):
        self.owner.vehicles = [Vehicle(type=VehicleType.car), Vehicle(type=VehicleType.bus)]
        db.flush()

        formset_class = inlineformset_factory(relation=Owner.vehicles, fields=("type",), session=db)
        formset = formset_class(instance=self.owner)

        soup = self.get_soup(formset.as_p())
        expected_soup = self.get_soup(
            "".join(
                [
                    '<input type="hidden" name="vehicles-TOTAL_FORMS" value="5" id="id_vehicles-TOTAL_FORMS" />',
                    '<input type="hidden" name="vehicles-INITIAL_FORMS" value="2" id="id_vehicles-INITIAL_FORMS" />',
                    '<input type="hidden" name="vehicles-MIN_NUM_FORMS" value="0" id="id_vehicles-MIN_NUM_FORMS" />',
                    '<input type="hidden" name="vehicles-MAX_NUM_FORMS" value="1000" id="id_vehicles-MAX_NUM_FORMS" />',
                    "<p>",
                    '  <label for="id_vehicles-0-type">Type:</label>',
                    '  <select id="id_vehicles-0-type" name="vehicles-0-type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car" selected>Car</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-0-DELETE">Delete:</label>',
                    '  <input id="id_vehicles-0-DELETE" name="vehicles-0-DELETE" type="checkbox" />',
                    f'  <input id="id_vehicles-0-id" name="vehicles-0-id" type="hidden" value="{self.owner.vehicles[0].id}" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-1-type">Type:</label>',
                    '  <select id="id_vehicles-1-type" name="vehicles-1-type">',
                    '    <option value="bus" selected>Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-1-DELETE">Delete:</label>',
                    '  <input id="id_vehicles-1-DELETE" name="vehicles-1-DELETE" type="checkbox" />',
                    f'  <input id="id_vehicles-1-id" name="vehicles-1-id" type="hidden" value="{self.owner.vehicles[1].id}" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-2-type">Type:</label>',
                    '  <select id="id_vehicles-2-type" name="vehicles-2-type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-2-DELETE">Delete:</label>',
                    '  <input id="id_vehicles-2-DELETE" name="vehicles-2-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-3-type">Type:</label>',
                    '  <select id="id_vehicles-3-type" name="vehicles-3-type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-3-DELETE">Delete:</label>',
                    '  <input id="id_vehicles-3-DELETE" name="vehicles-3-DELETE" type="checkbox" />',
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-4-type">Type:</label>',
                    '  <select id="id_vehicles-4-type" name="vehicles-4-type">',
                    '    <option value="bus">Bus</option>',
                    '    <option value="car">Car</option>',
                    "  </select>",
                    "</p>",
                    "<p>",
                    '  <label for="id_vehicles-4-DELETE">Delete:</label>',
                    '  <input id="id_vehicles-4-DELETE" name="vehicles-4-DELETE" type="checkbox" />',
                    "</p>",
                ]
            )
        )
        self.assertHTMLEqual(soup.prettify(), expected_soup.prettify())



"""Django detail view things for sqlalchemy."""

from django.core.exceptions import ImproperlyConfigured
from django.views.generic.base import TemplateResponseMixin, View

from .base import BaseSingleObjectMixin


class SingleObjectMixin(BaseSingleObjectMixin):
    """Provide the ability to retrieve a single object for further
    manipulation."""

    def get_context_object_name(self, obj):
        """Get the name to use for the object."""
        if self.context_object_name:
            return self.context_object_name

        model = self.get_model()
        return model.__name__.lower()

    def get_context_data(self, **kwargs):
        """Insert the single object into the context dict."""
        context = {}
        if self.object:
            context["object"] = self.object
            context_object_name = self.get_context_object_name(self.object)
            if context_object_name:
                context[context_object_name] = self.object
        context.update(kwargs)
        return super().get_context_data(**context)


class BaseDetailView(SingleObjectMixin, View):
    """A base view for displaying a single object."""

    def get(self, request, *args, **kwargs):
        """Handles GET on detail view."""
        self.object = self.get_object()
        context = self.get_context_data(object=self.object)
        return self.render_to_response(context)


class SingleObjectTemplateResponseMixin(TemplateResponseMixin):
    """Provide the ability to retrieve template names."""

    template_name_field = None
    template_name_suffix = "_detail"

    def get_template_names(self):
        """Returns template names for detail view."""
        try:
            names = super().get_template_names()
        except ImproperlyConfigured as e:
            names = []

            if self.object and self.template_name_field:
                name = getattr(self.object, self.template_name_field, None)
                if name:
                    names.insert(0, name)

            if hasattr(self, "get_model_template_name"):
                names.append(self.get_model_template_name())

            if not names:
                raise e.with_traceback(None)

        return names


class DetailView(SingleObjectTemplateResponseMixin, BaseDetailView):
    """Render a "detail" view of an object.

    By default this is a model instance looked up from `self.queryset`,
    but the view will support display of *any* object by overriding
    `self.get_object()`.
    """



from django_sorcery.db import signals  # noqa

from ..base import TestCase
from ..testapp.models import CompositePkModel, Owner, OwnerQuery, db


class TestSession(TestCase):
    def setUp(self):
        super().setUp()
        Owner.query.delete()
        db.commit()
        db.remove()

        signals.before_flush.connect(self._before_flush, weak=False)
        signals.after_flush.connect(self._after_flush, weak=False)
        signals.before_commit.connect(self._before_commit, weak=False)
        signals.after_commit.connect(self._after_commit, weak=False)

        self.before_flush_signal_called = False
        self.after_flush_signal_called = False
        self.before_commit_called = False
        self.after_commit_called = False
        self.models_committed = None
        self.models_deleted = None

    def _before_flush(self, session, **kwargs):
        self.before_flush_signal_called = True

    def _after_flush(self, session, **kwargs):
        self.after_flush_signal_called = True

    def _before_commit(self, session, **kwargs):
        self.before_commit_called = True

    def _after_commit(self, session, **kwargs):
        self.after_commit_called = True
        self.models_committed = session.models_committed.copy()
        self.models_deleted = session.models_deleted.copy()

    def tearDown(self):
        super().tearDown()
        signals.before_flush.disconnect(self._before_flush)
        signals.after_flush.disconnect(self._after_flush)
        signals.before_commit.disconnect(self._before_commit)
        signals.after_commit.disconnect(self._after_commit)
        Owner.query.delete()
        db.commit()

    def test_session_signals(self):
        db.add(Owner(first_name="Joe", last_name="Smith"))
        db.flush()

        self.assertTrue(self.before_flush_signal_called)
        self.assertTrue(self.after_flush_signal_called)

    def test_query_class_usage(self):
        query = db.query(Owner)
        self.assertIsInstance(query, OwnerQuery)

        query = db.query(CompositePkModel)
        self.assertIsInstance(query, db.query_class)

    def test_commit_signals(self):
        with db.atomic():
            db.add(Owner(first_name="Joe", last_name="Smith"))

        self.assertFalse(self.before_commit_called)
        self.assertFalse(self.after_commit_called)

        db.commit()

        self.assertTrue(self.before_commit_called)
        self.assertTrue(self.after_commit_called)

    def test_tracking(self):
        owner = Owner(first_name="Joe", last_name="Smith")
        db.add(owner)
        db.commit()

        self.assertEqual(set(self.models_committed), {owner})
        self.assertEqual(set(self.models_deleted), set())

        owner.first_name = "Michael"
        db.commit()

        self.assertEqual(set(self.models_committed), {owner})
        self.assertEqual(set(self.models_deleted), set())

        db.delete(owner)
        db.commit()

        self.assertEqual(set(self.models_committed), set())
        self.assertEqual(set(self.models_deleted), {owner})

        owner = Owner(first_name="Joe", last_name="Smith")
        db.add(owner)
        db.flush()
        db.delete(owner)
        db.commit()

        self.assertEqual(set(self.models_committed), set())
        self.assertEqual(set(self.models_deleted), {owner})



import sqlalchemy as sa

from .. import models_backpop
from ..base import TestCase
from ..minimalapp import models


class TestSQLAlchemyRelationships(TestCase):
    def tearDown(self):
        super().setUp()
        self.addCleanup(models.db.rollback)
        self.addCleanup(models.db.remove)
        self.addCleanup(models_backpop.db.rollback)
        self.addCleanup(models_backpop.db.remove)

    def _test_relations(self, module):
        self.assertTrue(hasattr(module.Asset, "_order_pk"))
        self.assertTrue(hasattr(module.Order, "_order_item_pk"))
        self.assertTrue(hasattr(module.Order, "_applicant_pk"))
        self.assertTrue(hasattr(module.Order, "_coapplicant_pk"))
        self.assertTrue(hasattr(module.Profile, "_customer_pk"))

        order = module.Order(applicant=module.Customer(profile=module.Profile()))
        order.coapplicant = module.Customer(profile=module.Profile())
        order.assets = [module.Asset(), module.Asset()]

        module.db.add(order)

        self.assertEqual(len(module.db.new), 7)

        module.db.flush()

        self.assertEqual(order.applicant.pk, order._applicant_pk)
        for item in order.assets:
            self.assertEqual(item._order_pk, order.pk)
            self.assertEqual(item.order, order)

        module.db.rollback()

        for _ in range(10):
            order = module.Order(applicant=module.Customer(profile=module.Profile()))
            order.coapplicant = module.Customer(profile=module.Profile())
            order.assets = [module.Asset(), module.Asset()]
            order.contacts = [module.Contact(), module.Contact()]

            other_order = module.Order()
            other_order.contacts.extend(order.contacts)

            module.db.add(order)

        module.db.flush()
        module.db.expire_all()

        association_table_name = module.Order.contacts.info["table_name"]
        self.assertIsNotNone(module.db.Model.metadata.tables.get(association_table_name))

        tbl = module.db.Model.metadata.tables.get(association_table_name)
        rows = module.db.execute(tbl.select()).fetchall()

        orderrows = {}
        contactrows = {}
        for row in rows:
            orderrows.setdefault(row.order_pk, set()).add(row.contact_pk)
            contactrows.setdefault(row.contact_pk, set()).add(row.order_pk)

        for order in module.Order.query:
            self.assertEqual(len(order.contacts), 2)
            for contract in order.contacts:
                self.assertIn(contract.pk, orderrows[order.pk])

        for contact in module.Contact.query:
            self.assertEqual(len(contact.orders), 2)
            for order in contact.orders:
                self.assertIn(order.pk, contactrows[contact.pk])

    def test_relations(self):
        self._test_relations(models)
        self._test_relations(models_backpop)

    def test_bad_many_to_many(self):
        with self.assertRaises(sa.exc.ArgumentError):

            class SuperDummy(models.db.Model):
                bad = models.db.ManyToMany("blah")



"""Helper functions for creating FormSet classes from SQLAlchemy models."""

from django.core.exceptions import ImproperlyConfigured
from django.forms import fields as djangofields
from django.forms.formsets import BaseFormSet, formset_factory
from django.forms.widgets import HiddenInput

from ..db import meta
from ..forms import ModelForm, modelform_factory


class BaseModelFormSet(BaseFormSet):
    """A ``FormSet`` for editing a queryset and/or adding new objects to it."""

    model = None
    session = None
    absolute_max = 2000

    # Set of fields that must be unique among forms of this set.
    unique_fields = set()

    def __init__(self, data=None, files=None, auto_id="id_%s", prefix=None, queryset=None, initial=None, **kwargs):
        self.session = kwargs.pop("session", self.session)
        self.queryset = queryset
        self.initial_extra = initial
        kwargs.update({"data": data, "files": files, "auto_id": auto_id, "prefix": prefix})
        super().__init__(**kwargs)

    def initial_form_count(self):
        """Return the number of forms that are required in this FormSet."""
        return super().initial_form_count() if (self.data or self.files) else len(self.get_queryset())

    def _existing_object(self, pk):
        info = meta.model_info(self.model)

        if not isinstance(pk, tuple):
            pk = (pk,)

        if not hasattr(self, "_object_dict"):
            self._object_dict = {info.get_key(o): o for o in self.get_queryset()}

        return self._object_dict.get(pk)

    def _construct_form(self, i, **kwargs):
        pk_required = i < self.initial_form_count()
        if pk_required:
            if self.is_bound:
                info = meta.model_info(self.model)
                pks = {}
                for name, pk_info in info.primary_keys.items():
                    pk_key = f"{self.add_prefix(i)}-{name}"
                    pk_val = self.data.get(pk_key)
                    pks[name] = pk_info.column.type.python_type(pk_val) if pk_val else None

                pk = info.primary_keys_from_dict(pks)

                kwargs["instance"] = self._existing_object(pk)
            else:
                kwargs["instance"] = self.get_queryset()[i]
        elif self.initial_extra:
            # Set initial values for extra forms
            try:
                kwargs["initial"] = self.initial_extra[i - self.initial_form_count()]
            except IndexError:
                pass
        kwargs["session"] = self.session
        return super()._construct_form(i, **kwargs)

    def add_fields(self, form, index):
        info = meta.model_info(self.model)

        if form.instance in self.session:
            for name in info.primary_keys:
                pk_field = djangofields.Field(initial=getattr(form.instance, name, None), widget=HiddenInput)
                form.fields[name] = pk_field

        super().add_fields(form, index)

    def get_queryset(self):
        """Returns a query for the model."""
        if not hasattr(self, "_queryset"):
            if self.queryset is not None:
                qs = self.queryset
            else:
                qs = self.session.query(self.model)[: self.absolute_max]

            if hasattr(qs, "all"):
                qs = qs.all()

            self._queryset = qs

        return self._queryset

    def delete_existing(self, obj, **kwargs):
        """Deletes an existing model instance."""
        self.session.delete(obj)

    def save(self, flush=True, **kwargs):
        """Save model instances for every form, adding and changing instances
        as necessary, and return the list of instances."""
        self.new_objects = []
        self.changed_objects = []
        self.deleted_objects = []
        saved_instances = []

        for form in self.extra_forms:
            if not form.has_changed():
                continue

            if self.can_delete and self._should_delete_form(form):
                continue

            obj = form.save(flush=flush)
            self.new_objects.append(obj)
            saved_instances.append(obj)

        for form in self.initial_forms:
            if form in self.deleted_forms:
                self.deleted_objects.append(form.instance)
                self.delete_existing(form.instance)
                continue

            elif form.has_changed():
                self.changed_objects.append(form.instance)

            obj = form.save(flush=flush)
            saved_instances.append(obj)

        return saved_instances

    save.alters_data = True


def modelformset_factory(
    model,
    form=ModelForm,
    formfield_callback=None,
    formset=BaseModelFormSet,
    extra=1,
    can_delete=False,
    can_order=False,
    max_num=None,
    fields=None,
    exclude=None,
    widgets=None,
    validate_max=False,
    localized_fields=None,
    labels=None,
    help_texts=None,
    error_messages=None,
    min_num=None,
    validate_min=False,
    field_classes=None,
    session=None,
):
    """Return a FormSet class for the given sqlalchemy model class."""
    _meta = getattr(form, "Meta", None)
    fields = getattr(_meta, "fields", fields)
    exclude = getattr(_meta, "exclude", exclude)
    if fields is None and exclude is None:
        raise ImproperlyConfigured(
            "Calling modelformset_factory without defining 'fields' or " "'exclude' explicitly is prohibited."
        )

    form = modelform_factory(
        model,
        form=form,
        fields=fields,
        exclude=exclude,
        formfield_callback=formfield_callback,
        widgets=widgets,
        localized_fields=localized_fields,
        labels=labels,
        help_texts=help_texts,
        error_messages=error_messages,
        field_classes=field_classes,
        session=session,
    )

    FormSet = formset_factory(
        form,
        formset,
        extra=extra,
        min_num=min_num,
        max_num=max_num,
        can_order=can_order,
        can_delete=can_delete,
        validate_min=validate_min,
        validate_max=validate_max,
    )
    class_name = f"{model.__name__}FormSet"
    return type(form)(str(class_name), (FormSet,), {"model": model, "session": session})



"""Metadata for sqlalchemy model relationships."""
from itertools import chain

import sqlalchemy as sa
from django.core.exceptions import ImproperlyConfigured


class relation_info:
    """A helper class that makes sqlalchemy relationship property inspection
    easier."""

    __slots__ = (
        "attribute",
        "direction",
        "field_kwargs",
        "foreign_keys",
        "local_remote_pairs",
        "local_remote_pairs_for_identity_key",
        "name",
        "parent_mapper",
        "parent_model",
        "parent_table",
        "related_mapper",
        "related_model",
        "related_table",
        "relationship",
        "uselist",
    )

    def __init__(self, relationship):
        self.relationship = relationship
        self.name = self.relationship.key
        self.parent_mapper = self.relationship.parent
        self.parent_model = self.relationship.parent.class_
        self.parent_table = self.relationship.parent.tables[0]
        self.related_mapper = self.relationship.mapper
        self.related_model = self.relationship.mapper.class_
        self.related_table = self.relationship.mapper.tables[0]
        self.attribute = getattr(self.parent_model, self.name)
        self.direction = self.relationship.direction
        self.foreign_keys = list(
            set(chain(self.relationship._calculated_foreign_keys, self.relationship._user_defined_foreign_keys))
        )
        self.uselist = self.relationship.uselist
        self.field_kwargs = {}
        if self.uselist:
            self.field_kwargs["required"] = False
        else:
            self.field_kwargs["required"] = not all(col.nullable for col in self.foreign_keys)

        self.local_remote_pairs = self.relationship.local_remote_pairs

        target_pk = list(self.relationship.target.primary_key)
        pairs = {v: k for k, v in self.local_remote_pairs}
        self.local_remote_pairs_for_identity_key = []
        try:
            # ensure local_remote pairs are of same order as remote pk
            for i in target_pk:
                self.local_remote_pairs_for_identity_key.append((pairs[i], i))
        except KeyError:
            # if relation is missing one of related pk columns
            # but actual constraint has it defined
            # attempt to deduce what is the missing pk column
            # by inspecting FK constraint on table object itself
            # this only happens for pretty bad table structure
            # where some columns need to be omitted from relation
            # since same column is used in multiple relations
            # for people reading this comment lesson should be
            # DO NOT USE COMPOSITE PKs
            parent_columns = set(pairs.values())
            target_columns = set(target_pk)
            matching_constraints = [
                i
                for i in [c for c in self.parent_table.constraints if isinstance(c, sa.ForeignKeyConstraint)]
                if parent_columns & {j.parent for j in i.elements} == parent_columns
                and target_columns & {j.column for j in i.elements} == target_columns
            ]

            if len(matching_constraints) == 1:
                pairs = {i.column: i.parent for i in matching_constraints[0].elements}
                for i in target_pk:
                    self.local_remote_pairs_for_identity_key.append((pairs[i], i))
            else:
                # if everything fails, return default pairs
                self.local_remote_pairs_for_identity_key = self.local_remote_pairs[:]

    def formfield(self, form_class=None, **kwargs):
        if kwargs.get("session") is None:
            raise ImproperlyConfigured("Creating a relation form field without session attribute prohibited")

        form_class = form_class or self.get_form_class()
        if form_class is not None:
            field_kwargs = self.field_kwargs.copy()
            field_kwargs.update(kwargs)
            return form_class(self.related_model, **field_kwargs)

    def get_form_class(self):
        from ...fields import ModelChoiceField, ModelMultipleChoiceField

        if self.uselist:
            return ModelMultipleChoiceField

        return ModelChoiceField

    def __repr__(self):
        return "<relation_info({}.{})>".format(self.parent_model.__name__, self.name)



from django.conf import settings
from django_sorcery.db.query import QueryProperty

from ..base import TestCase
from ..testapp.models import (
    CompositePkModel,
    Owner,
    Point,
    Vehicle,
    VehicleType,
    Vertex,
    db,
)


class TestQuery(TestCase):
    def setUp(self):
        super().setUp()
        self.owner = Owner(first_name="Test 1", last_name="Owner 1")
        vehicle = Vehicle(name="used", is_used=True, type=VehicleType.car, owner=self.owner)
        vertex = Vertex(start=Point(x=1, y=2), end=Point(x=3, y=4))
        db.add(self.owner)
        db.add(vehicle)
        db.add(vertex)
        db.add(CompositePkModel(id=1, pk=1, active=True, name="Test-1-1"))
        db.flush()
        self.owner_id = self.owner.id
        self.vehicle_id = vehicle.id
        self.vertex_id = vertex.pk
        db.expire_all()

    def tearDown(self):
        super().tearDown()
        db.rollback()
        db.remove()

    def test_query_get_regular(self):
        owner = Owner.query.get(self.owner_id)
        self.assertEqual(owner.id, self.owner_id)

    def test_query_get_with_kwargs(self):
        owner = Owner.query.get(id=self.owner_id)
        self.assertEqual(owner.id, self.owner_id)

    def test_query_get_with_kwargs_composite(self):
        obj = CompositePkModel.query.get(id=1, pk=1)
        self.assertEqual(obj.name, "Test-1-1")

    def test_query_get_with_kwargs_tuple(self):
        obj = CompositePkModel.query.get((1, 1))
        self.assertEqual(obj.name, "Test-1-1")

    def test_query_kwarg_none(self):
        self.assertIsNone(CompositePkModel.query.get(id=1))

    def test_query_order_by_django_style(self):
        owner1 = Owner(first_name="AAAA 1", last_name="AAAA 1")
        owner2 = Owner(first_name="ZZZZ 2", last_name="ZZZZ 1")
        db.add_all([owner1, owner2])
        db.flush()

        owners = Owner.objects.order_by("-first_name").all()

        self.assertListEqual([owner2, self.owner, owner1], owners)

        owners = Owner.objects.order_by("+first_name", "-last_name").all()
        self.assertListEqual([owner1, self.owner, owner2], owners)

    def test_query_filter_default_equality(self):
        obj = Owner.query.filter(first_name="Test 1").first()
        self.assertEqual(obj.id, self.owner_id)

    def test_query_filter_lookup(self):
        obj = Owner.query.filter(first_name__istartswith="test").first()
        self.assertEqual(obj.id, self.owner_id)

    def test_query_filter_relation_with_lookup(self):
        obj = Vehicle.query.filter(owner__first_name__istartswith="test").first()
        self.assertEqual(obj.id, self.vehicle_id)

    def test_query_filter_relation_with_instance(self):
        obj = Vehicle.query.filter(owner=Owner.objects.get(id=self.owner_id)).first()
        self.assertEqual(obj.id, self.vehicle_id)

    def test_query_filter_composite_with_lookup(self):
        obj = Vertex.query.filter(start__x__gte=1).first()
        self.assertEqual(obj.pk, self.vertex_id)

    def test_query_filter_composite_with_instance(self):
        obj = Vertex.query.filter(start=Vertex.query.get(self.vertex_id).start).first()
        self.assertEqual(obj.pk, self.vertex_id)


class TestQueryProperty(TestCase):
    def setUp(self):
        super().setUp()
        db.add_all(
            [
                Vehicle(name="used", is_used=True, type=VehicleType.car),
                Vehicle(name="new", is_used=False, type=VehicleType.car),
            ]
        )
        db.flush()

    def test_options_filter_and_filter_by(self):
        class Dummy:
            vehicles = QueryProperty(db, Vehicle).options(db.joinedload(Vehicle.owner))
            used_vehicles = QueryProperty(db, Vehicle, Vehicle.is_used.is_(True))
            new_vehicles = QueryProperty(db, Vehicle, is_used=False)

        dummy = Dummy()

        self.assertEqual(len(dummy.vehicles.all()), 2)
        self.assertEqual(dummy.used_vehicles.count(), 1)
        self.assertEqual(dummy.new_vehicles.count(), 1)

    def test_bad_attr(self):
        with self.assertRaises(AttributeError) as ctx:
            QueryProperty(db, Vehicle).dummy(db.joinedload(Vehicle.owner))

        self.assertEqual(
            ctx.exception.args,
            (
                f"<QueryProperty db=<SQLAlchemy engine=postgresql://postgres:***@{settings.DB_URL.host}/test>, "
                "model='Vehicle'> "
                "object has no attribute 'dummy'",
            ),
        )

    def test_no_model(self):
        class Dummy:
            vehicles = QueryProperty(db, None)

        with self.assertRaises(AttributeError) as ctx:
            Dummy.vehicles

        self.assertEqual(
            ctx.exception.args,
            (
                "Cannot access QueryProperty when not bound to a model. You can explicitly instantiate descriptor with "
                "model class - `db.queryproperty(Model)`.",
            ),
        )

    def test_stupid(self):
        qp = QueryProperty(db, Vehicle)

        class Dummy:
            vehicles = qp

        qp.model = object

        self.assertIsNone(Dummy.vehicles)



"""Django-esque field metadata and interface providers."""
import datetime
import decimal
import enum
from contextlib import suppress

import sqlalchemy as sa
from dateutil.parser import parse
from django import forms as djangoforms
from django.conf import settings
from django.core import validators as djangovalidators
from django.core.exceptions import ValidationError
from django.db.models import fields as djangomodelfields
from django.forms import fields as djangofields
from django.utils import timezone
from django.utils.text import capfirst

from ... import fields as sorceryfields
from ...utils import sanitize_separators


def _make_naive(value):
    if settings.USE_TZ and timezone.is_aware(value):
        default_timezone = timezone.get_default_timezone()
        value = timezone.make_naive(value, default_timezone)
    return value


class column_info:
    """A helper class that makes sqlalchemy property and column inspection
    easier."""

    default_form_class = None
    default_error_messages = djangomodelfields.Field.default_error_messages
    is_relation = False

    __slots__ = (
        "_coercer",
        "attname",
        "attribute",
        "choices",
        "column",
        "label",
        "default",
        "empty_values",
        "error_messages",
        "field_kwargs",
        "form_class",
        "help_text",
        "unique",
        "name",
        "null",
        "parent",
        "parent_model",
        "property",
        "required",
        "validators",
        "widget",
    )

    def __new__(cls, *args, **kwargs):
        args = list(args)
        column = kwargs.pop("column", None)

        if args:
            column = args.pop(0)

        column_info_mapping = getattr(settings, "DJANGO_SORCERY", {}).get("column_info_mapping", COLUMN_INFO_MAPPING)

        override_cls = None
        for base in column.type.__class__.mro():
            if base in column_info_mapping:
                override_cls = column_info_mapping.get(base, cls)

            enum_class = getattr(column.type, "enum_class", object) or object
            for sub in enum_class.mro():
                if (base, sub) in column_info_mapping:
                    override_cls = column_info_mapping.get((base, sub), cls)
                    break

            if override_cls:
                break

        _cls = override_cls or cls
        return super().__new__(_cls)

    def __init__(self, column, prop=None, parent=None, name=None):
        self.property = prop
        self.column = column
        self.parent = parent
        self._coercer = None
        self.name = name or (self.property.key if self.property is not None else self.column.key)
        self.attname = self.name

        self.validators = self.column.info.get("validators") or []
        self.null = not self.column.primary_key and self.column.nullable
        self.required = self.column.info.get("required", not self.column.nullable)
        self.unique = self.column.unique

        self.parent_model = self.property.parent.class_ if self.property else None
        self.attribute = getattr(self.parent_model, self.property.key) if self.parent_model else None
        self.help_text = self.column.doc
        self.form_class = self.column.info.get("form_class") or self.default_form_class
        self.empty_values = self.column.info.get("empty_values") or getattr(
            self.form_class, "empty_values", djangovalidators.EMPTY_VALUES
        )
        self.default = getattr(self.column.default, "arg", None)
        self.choices = getattr(self.column.type, "enum_class", None) or getattr(self.column.type, "enums", None)
        self.widget = self.column.info.get("widget_class")

        self.error_messages = {}
        for c in reversed(self.__class__.mro()):
            self.error_messages.update(getattr(c, "default_error_messages", {}))
        self.error_messages.update(column.info.get("error_messages") or {})

        self.label = self.column.info.get("label") or (capfirst(" ".join(self.name.split("_"))) if self.name else None)

        self.field_kwargs = {"required": self.required, "validators": self.validators, "help_text": self.help_text}
        if self.default and not callable(self.default):
            self.field_kwargs["initial"] = self.default

        if self.label:
            self.field_kwargs["label"] = self.label

        if self.widget:
            self.field_kwargs["widget"] = self.widget

    def __repr__(self):
        return "<{!s}({!s}.{!s}){!s}>".format(
            self.__class__.__name__,
            self.parent.model_class.__name__ if self.parent else "<None>",
            self.name or "<None>",
            " pk" if self.column.primary_key else "",
        )

    @property
    def coercer(self):
        """Form field to be used to coerce data types."""
        if not self._coercer:
            self._coercer = self.formfield(localize=True) or djangofields.Field(localize=True)
        return self._coercer

    def formfield(self, form_class=None, **kwargs):
        """Returns the form field for the field."""
        form_class = form_class or self.form_class

        if form_class is not None:
            field_kwargs = self.field_kwargs.copy()
            field_kwargs.update(kwargs)
            return form_class(**field_kwargs)

    def to_python(self, value):
        """Convert input value to appropriate python object."""
        return self.coercer.to_python(value)

    def clean(self, value, instance):
        """Convert the value's type and run validation.

        Validation errors from to_python() and validate() are
        propagated. Return the correct value if no error is raised.
        """
        value = self.to_python(value)
        self.validate(value, instance)
        self.run_validators(value)
        return value

    def validate(self, value, instance):
        """Validate value and raise ValidationError if necessary."""
        getattr(instance, "clean_" + self.name, bool)()

    def run_validators(self, value):
        """Run field's validators and raise ValidationError if necessary."""
        if value in self.empty_values:
            return

        errors = []
        for v in self.validators:
            try:
                v(value)
            except ValidationError as e:
                errors.extend(e.error_list)

        if errors:
            raise ValidationError(errors)


class string_column_info(column_info):
    """Provides meta info for string columns."""

    default_form_class = djangofields.CharField

    def __init__(self, column, prop=None, parent=None, name=None):
        super().__init__(column, prop, parent, name)
        self.field_kwargs["max_length"] = self.column.type.length

    def to_python(self, value):
        if value is None:
            return value
        return str(value).strip()


class text_column_info(string_column_info):
    """Provides meta info for text columns."""

    def __init__(self, column, prop=None, parent=None, name=None):
        super().__init__(column, prop, parent, name)
        self.widget = self.column.info.get("widget_class") or djangoforms.Textarea
        self.field_kwargs["widget"] = self.widget


class choice_column_info(column_info):
    """Provides meta info for enum columns with simple choices."""

    default_form_class = djangofields.TypedChoiceField

    def __init__(self, column, prop=None, parent=None, name=None):
        super().__init__(column, prop, parent, name)
        self.field_kwargs["choices"] = [(x, x) for x in self.choices]
        # Many of the subclass-specific formfield arguments (min_value,
        # max_value) don't apply for choice fields, so be sure to only pass
        # the values that TypedChoiceField will understand.
        for k in list(self.field_kwargs):
            if k not in (
                "choices",
                "coerce",
                "disabled",
                "empty_value",
                "enum_class",
                "error_messages",
                "help_text",
                "initial",
                "label",
                "required",
                "show_hidden_initial",
                "validators",
                "widget",
            ):
                del self.field_kwargs[k]  # pragma: nocover

    def to_python(self, value):
        if value is None:
            return value

        with suppress(TypeError, ValueError):
            parsed = type(next(iter(self.choices)))(value)
            if parsed in self.choices:
                return parsed

        parsed = str(value).strip()
        parsed = self.coercer.to_python(parsed)
        if parsed in self.coercer.empty_values:
            return None

        raise ValidationError("%(value)r is not a valid choice.", code="invalid", params={"value": str(value)})


class enum_column_info(choice_column_info):
    """Provides meta info for enum columns with Enum choices."""

    default_form_class = sorceryfields.EnumField

    def __init__(self, column, prop=None, parent=None, name=None):
        super().__init__(column, prop, parent, name)
        self.field_kwargs["choices"] = self.choices

    def to_python(self, value):
        if value is None:
            return value

        with suppress(TypeError, KeyError, ValueError):
            return self.choices[value]
        with suppress(TypeError, KeyError, ValueError):
            return self.choices(value)
        with suppress(TypeError, AttributeError):
            return getattr(self.choices, value)

        return self.coercer.to_python(value)


class numeric_column_info(column_info):
    """Provides meta info for numeric columns."""

    __slots__ = ("max_digits", "decimal_places")

    default_form_class = djangofields.DecimalField

    def __init__(self, column, prop=None, parent=None, name=None):
        super().__init__(column, prop, parent, name)
        self.max_digits = self.column.type.precision
        self.decimal_places = self.column.type.scale
        if self.column.type.python_type == decimal.Decimal:
            if self.max_digits is not None:
                self.field_kwargs["max_digits"] = self.max_digits
            if self.decimal_places is not None:
                self.field_kwargs["decimal_places"] = self.decimal_places

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, decimal.Decimal):
            return value
        if isinstance(value, float):
            value = decimal.Context(prec=self.max_digits).create_decimal_from_float(value)
            return value.to_integral() if value == value.to_integral() else value.normalize()
        if isinstance(value, int):
            return decimal.Decimal(value)

        parsed = sanitize_separators(str(value).strip())
        return self.coercer.to_python(parsed)


class boolean_column_info(column_info):
    """Provides meta info for boolean columns."""

    def __init__(self, column, prop=None, parent=None, name=None):
        super().__init__(column, prop, parent, name)
        if not self.form_class:
            self.form_class = djangofields.NullBooleanField if self.null else djangofields.BooleanField

    def to_python(self, value):
        if value is None:
            return value
        if value in (True, False):
            return bool(value)
        if value in ("t", "T"):
            return True
        if value in ("f", "F"):
            return False
        return self.coercer.to_python(value)


class date_column_info(column_info):
    """Provides meta info for date columns."""

    default_form_class = djangofields.DateField

    @property
    def coercer(self):
        coercer = super().coercer
        coercer.input_formats = settings.DATE_INPUT_FORMATS
        return coercer

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, datetime.datetime):
            return _make_naive(value).date()
        if isinstance(value, datetime.date):
            return value

        parsed = str(value).strip()
        with suppress(ValueError):
            return _make_naive(datetime.datetime.fromtimestamp(float(parsed))).date()
        with suppress(ValueError):
            return _make_naive(parse(parsed)).date()

        return _make_naive(self.coercer.to_python(parsed)).date()


class datetime_column_info(column_info):
    """Provides meta info for datetime columns."""

    default_form_class = djangofields.DateTimeField

    @property
    def coercer(self):
        coercer = super().coercer
        coercer.input_formats = settings.DATETIME_INPUT_FORMATS
        return coercer

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, datetime.datetime):
            return _make_naive(value)
        if isinstance(value, datetime.date):
            return _make_naive(datetime.datetime(value.year, value.month, value.day))

        parsed = str(value).strip()
        with suppress(ValueError):
            return _make_naive(datetime.datetime.fromtimestamp(float(parsed)))

        with suppress(ValueError):
            return _make_naive(parse(parsed))

        return _make_naive(self.coercer.to_python(parsed))


class float_column_info(column_info):
    """Provides meta info for float columns."""

    default_form_class = djangofields.FloatField

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, float):
            return value

        parsed = str(value).strip()
        return self.coercer.to_python(parsed)


class integer_column_info(column_info):
    """Provides meta info for integer columns."""

    default_form_class = djangofields.IntegerField

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, int):
            return value

        parsed = str(value).strip()
        return self.coercer.to_python(parsed)


class interval_column_info(column_info):
    """Provides meta info for interval columns."""

    default_form_class = djangofields.DurationField

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, datetime.timedelta):
            return value

        parsed = str(value).strip()
        return self.coercer.to_python(parsed)


class time_column_info(column_info):
    """Provides meta info for time columns."""

    default_form_class = djangofields.TimeField

    def to_python(self, value):
        if value is None:
            return value
        if isinstance(value, datetime.time):
            return value
        if isinstance(value, datetime.datetime):
            return value.time()

        parsed = str(value).strip()
        return self.coercer.to_python(parsed)


COLUMN_INFO_MAPPING = {
    (sa.sql.sqltypes.Enum, enum.Enum): enum_column_info,
    (sa.sql.sqltypes.Enum, object): choice_column_info,
    sa.sql.sqltypes.String: string_column_info,
    sa.sql.sqltypes.Text: text_column_info,
    sa.sql.sqltypes.Numeric: numeric_column_info,
    sa.sql.sqltypes.Float: float_column_info,
    sa.sql.sqltypes.Integer: integer_column_info,
    sa.sql.sqltypes.Boolean: boolean_column_info,
    sa.sql.sqltypes.Date: date_column_info,
    sa.sql.sqltypes.DateTime: datetime_column_info,
    sa.sql.sqltypes.Interval: interval_column_info,
    sa.sql.sqltypes.Time: time_column_info,
}



"""Testing utilities."""

from sqlalchemy import event
from sqlalchemy.engine import Engine

from .db import databases


class CommitException(BaseException):
    """Raised when a commit happens during testing."""


class Transact:
    """Helper context manager for handling transactions during tests. Perfect
    for testing a single unit of work.

    For testing sqlalchemy apps the common pattern is to usually, start
    a transaction or savepoint, run the tests and once the test is done
    rollback. Additionally it also prevents commits to the database so
    that every test gets the same database state.
    """

    @staticmethod
    def _commit(connection, *args, **kwargs):
        raise CommitException("Commits are not allowed")

    def start(self):
        """Starts transaction management and hooks commit events."""
        self.hook()

    def stop(self):
        """Stops transaction management, rolls back transactions and unhooks
        commit events."""
        self.end()
        self.unhook()

    def hook(self):
        """Hooks commit events to prevent commits."""
        if not event.contains(Engine, "commit", self._commit):
            event.listen(Engine, "commit", self._commit)
        if not event.contains(Engine, "commit_twophase", self._commit):
            event.listen(Engine, "commit_twophase", self._commit)

    def unhook(self):
        """Unhooks commit events."""
        if event.contains(Engine, "commit", self._commit):
            event.remove(Engine, "commit", self._commit)
        if event.contains(Engine, "commit_twophase", self._commit):
            event.remove(Engine, "commit_twophase", self._commit)

    def end(self):
        """Rolls back transaction and removes session."""
        databases.rollback()
        databases.remove()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



from django.http import HttpResponse
from django.test import RequestFactory, override_settings
from django_sorcery.db.profiler import (
    SQLAlchemyProfiler,
    SQLAlchemyProfilingMiddleware,
)

from ..base import TestCase
from ..testapp.models import Business, Owner, db


def get_response(request):
    db.add(Owner(first_name="foo", last_name="bar"))
    db.flush()
    db.rollback()
    db.remove()
    return HttpResponse()


class TestSQLAlchemyProfilingMiddleware(TestCase):
    @override_settings(DEBUG=True)
    def test_profiler_debug(self):
        m = SQLAlchemyProfilingMiddleware(get_response)

        response = m(RequestFactory().get("/"))

        self.assertEqual(response["X-SA-Insert"], "1")

    @override_settings(DEBUG=False)
    def test_profiler_no_debug(self):
        m = SQLAlchemyProfilingMiddleware(get_response)

        response = m(RequestFactory().get("/"))

        self.assertNotIn("X-SA-Insert", response)

    def test_log_results(self):
        with override_settings(DEBUG=True):
            self.assertTrue(SQLAlchemyProfilingMiddleware(get_response).log_results)
        with override_settings(DEBUG=False):
            self.assertFalse(SQLAlchemyProfilingMiddleware(get_response).log_results)

    def test_header_results(self):
        with override_settings(DEBUG=True):
            self.assertTrue(SQLAlchemyProfilingMiddleware(get_response).log_results)
        with override_settings(DEBUG=False):
            self.assertFalse(SQLAlchemyProfilingMiddleware(get_response).log_results)


class TestProfiler(TestCase):
    def test_profiler(self):
        profiler = SQLAlchemyProfiler(exclude=["business"])

        with profiler:
            db.add(Owner(first_name="foo", last_name="bar"))
            db.flush()
            Owner.objects.all()
            Business.objects.all()
            db.rollback()
            db.remove()

        self.assertDictEqual(
            profiler.counts,
            {
                "begin": 1,
                "engine_connect": 1,
                "execute": 2,
                "insert": 1,
                "pool_checkin": 1,
                "pool_checkout": 1,
                "pool_reset": 1,
                "rollback": 1,
                "select": 1,
            },
        )

        insert_query, select_query = profiler.queries

        self.assertTrue(insert_query.statement.lower().startswith("insert into owner"))
        self.assertTrue(insert_query.parameters, [{"first_name": "foo", "last_name": "bar"}])

        self.assertTrue(select_query.statement.lower().startswith("select owner.id"))
        self.assertTrue(select_query.parameters, [{}])



"""Alembic Django command things."""
import os
from collections import OrderedDict, namedtuple

import alembic
import alembic.config
from django.core.management.base import BaseCommand, CommandError
from django.utils.functional import cached_property
from sqlalchemy.orm import configure_mappers

from ..db import alembic as sorcery_alembic, databases, meta, signals
from ..db.alembic import include_object, process_revision_directives


SORCERY_ALEMBIC_CONFIG_FOLDER = os.path.abspath(os.path.dirname(sorcery_alembic.__file__))


AlembicAppConfig = namedtuple("AlembicAppConfig", ["name", "config", "script", "db", "app", "version_path", "tables"])


class AlembicCommand(BaseCommand):
    """Base alembic django command."""

    @cached_property
    def sorcery_apps(self):
        """All sorcery apps and their alembic configs."""
        configs = OrderedDict()
        for db in databases.values():
            table_class_map = {model.__table__: model for model in db.models_registry if hasattr(model, "__table__")}
            for table in db.metadata.sorted_tables:
                model = table_class_map.get(table)
                if model:
                    info = meta.model_info(model)
                    app = info.app_config
                    if app:
                        path = self.get_app_version_path(app)
                        if os.path.exists(path):
                            config = self.get_app_config(app, db)
                            appconfig = AlembicAppConfig(
                                name=app.label,
                                config=config,
                                db=db,
                                script=self.get_config_script(config),
                                version_path=path,
                                app=app,
                                tables=[],
                            )
                            configs.setdefault(app.label, appconfig).tables.append(table)
        for app in configs.values():
            signals.alembic_app_created.send(app.app)
            signals.alembic_config_created.send(app.config)
        return configs

    def get_app_config(self, app, db):
        """Return alembic config for an app."""
        # TODO: read these from django db settings
        version_table = (
            getattr(app, "version_table", None)
            or db.kwargs.get("version_table")
            or "alembic_version_%s" % app.label.lower().replace(".", "_")
        )

        max_length = db.engine.dialect.max_identifier_length
        if max_length and len(version_table) >= max_length:
            raise CommandError(
                "'{name}' is {length} characters long which is an invalid identifier "
                "in {dialect!r} as its max idenfier length is {max_length}".format(
                    name=version_table, dialect=db.engine.dialect.name, length=len(version_table), max_length=max_length
                )
            )

        version_table_schema = getattr(app, "version_table_schema", None) or db.kwargs.get("version_table_schema")

        config = alembic.config.Config(output_buffer=self.stdout, stdout=self.stdout)
        config.set_main_option("script_location", SORCERY_ALEMBIC_CONFIG_FOLDER)
        config.set_main_option("version_locations", self.get_app_version_path(app))
        config.set_main_option("version_table", version_table)
        if version_table_schema and db.engine.dialect.name != "sqlite":
            config.set_main_option("version_table_schema", version_table_schema)
        return config

    def get_config_script(self, config):
        """Returns the alembic script directory for the config."""
        return alembic.script.ScriptDirectory.from_config(config)

    def lookup_app(self, app_label):
        """Looks up an app's alembic config."""
        if app_label not in self.sorcery_apps:
            raise CommandError("App '%s' could not be found. Is it in INSTALLED_APPS?" % app_label)

        return self.sorcery_apps[app_label]

    def get_app_version_path(self, app):
        """Returns the default migration directory location of al app."""
        return os.path.join(app.path, "migrations")

    def get_common_config(self, context):
        """Common alembic configuration."""
        config = context.config
        return {
            "include_object": include_object,
            "process_revision_directives": process_revision_directives,
            "version_table": config.get_main_option("version_table"),
            "version_table_schema": config.get_main_option("version_table_schema"),
        }

    def run_env(self, context, appconfig):
        """Executes an alembic context, just like the env.py file of
        alembic."""
        configure_mappers()
        try:
            if context.is_offline_mode():
                self.run_migrations_offline(context, appconfig)
            else:
                self.run_migrations_online(context, appconfig)
        except alembic.util.exc.CommandError as e:
            raise CommandError(str(e))

    def run_migrations_online(self, context, appconfig):
        """Executes an online alembic context."""
        with appconfig.db.engine.connect() as connection:
            context.configure(
                connection=connection, target_metadata=appconfig.db.metadata, **self.get_common_config(context)
            )

            with context.begin_transaction():
                context.run_migrations()

    def run_migrations_offline(self, context, appconfig):
        """Executes an offline alembic context."""
        context.configure(
            url=appconfig.db.url,
            literal_binds=True,
            target_metadata=appconfig.db.metadata,
            **self.get_common_config(context),
        )
        with context.begin_transaction():
            context.run_migrations()



"""Django middleware support for sqlalchemy."""
import logging

from . import databases
from .signals import all_signals


before_middleware_request = all_signals.signal("before_middleware_request")
after_middleware_response = all_signals.signal("after_middleware_response")

logger = logging.getLogger(__name__)


class BaseMiddleware:
    """Base middleware implementation that supports unit of work per request
    for django."""

    logger = logger

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.process_request(request)
        if response is not None:
            return self.return_response(request, response)

        response = self.get_response(request)
        return self.process_response(request, response)

    def process_request(self, request):
        """Hook for adding arbitrary logic to request processing."""
        before_middleware_request.send(self.__class__, middleware=self, request=request)

    def process_response(self, request, response):
        """Commits or rollbacks scoped sessions depending on status code then
        removes them."""
        if response.status_code >= 400:
            self.rollback(request=request, response=response)
            return self.return_response(request, response)

        if request.method not in {"PUT", "POST", "PATCH", "GET", "DELETE"}:
            self.rollback(request=request, response=response)
            return self.return_response(request, response)

        try:
            self.flush(request=request, response=response)
            self.commit(request=request, response=response)
        except Exception:
            self.logger.error("Error during flush or commit")
            self.rollback(request=request, response=response)
            self.return_response(request, response)
            raise

        return self.return_response(request, response)

    def return_response(self, request, response):
        """Hook for adding arbitrary logic to response processing."""
        self.remove(request=request, response=response)

        after_middleware_response.send(self.__class__, middleware=self, request=request, response=response)

        return response


class SQLAlchemyDBMiddleware(BaseMiddleware):
    """A base SQLAlchemy db middleware.

    Used by SQLAlchemy to provide a default middleware for a single db,
    it will first try to flush and if successfull, proceed with commit.
    If there are any errors during flush, will issue a rollback.
    """

    db = None

    def rollback(self, request, response):
        """Rolls back current scoped session."""
        self.db.rollback()

    def flush(self, request, response):
        """Flushes current scoped session."""
        self.db.flush()

    def commit(self, request, response):
        """Commits current scoped session."""
        self.db.commit()

    def remove(self, request, response):
        """Removes current scoped session."""
        self.db.remove()


class SQLAlchemyMiddleware(SQLAlchemyDBMiddleware):
    """A sqlalchemy middleware that manages all the dbs configured and
    initialized.

    it will first try to flush all the configured and initialized
    SQLAlchemy instances and if successfull, proceed with commit. If
    there are any errors during flush, all transactions will be rolled
    back.
    """

    db = databases



from django_sorcery.db import databases


db = databases.get("terrible")


class Foo(db.Model):
    id1 = db.Column(db.Integer(), primary_key=True)
    id2 = db.Column(db.Integer(), primary_key=True)

    parent_id2 = db.Column(db.Integer())

    partial_parent = db.relationship(
        "Foo",
        foreign_keys=[parent_id2],
        primaryjoin=db.and_(id1 == id1, db.remote(id2) == db.foreign(parent_id2)),
        uselist=False,
    )

    class Meta:
        table_args = db.args(db.ForeignKeyConstraint(["id1", "parent_id2"], ["foo.id1", "foo.id2"], use_alter=True))


db.create_all()



"""Metadata for sqlalchemy models."""
from collections import OrderedDict, namedtuple
from functools import partial
from itertools import chain

import inflect
import sqlalchemy as sa
from django.apps import apps
from django.core.exceptions import FieldDoesNotExist, ValidationError

from ...exceptions import NestedValidationError
from ...validators import ValidationRunner
from .base import model_info_meta
from .column import column_info
from .composite import composite_info
from .relations import relation_info


Identity = namedtuple("Key", ["model", "pk"])
inflector = inflect.engine()


class model_info(metaclass=model_info_meta):
    """A helper class that makes sqlalchemy model inspection easier."""

    __slots__ = (
        "field_names",
        "app_label",
        "composites",
        "concrete_fields",
        "fields",
        "local_fields",
        "mapper",
        "label",
        "label_lower",
        "column_properties",
        "model",
        "model_class",
        "model_name",
        "object_name",
        "opts",
        "ordering",
        "primary_keys",
        "private_fields",
        "properties",
        "relationships",
        "unique_together",
        "verbose_name",
        "verbose_name_plural",
    )

    def __init__(self, model):
        self.model_class = self.model = model
        self.mapper = sa.inspect(model)
        self.field_names = ()
        self.properties = OrderedDict()
        self.composites = OrderedDict()
        self.relationships = OrderedDict()
        self.primary_keys = OrderedDict()
        self.private_fields = ()
        self.local_fields = ()
        self.concrete_fields = ()

        self.opts = getattr(model, "Meta", None)

        self.object_name = getattr(self.opts, "object_name", self.model_class.__name__)
        self.model_name = getattr(self.opts, "model_name", self.object_name.lower())
        self.verbose_name = getattr(self.opts, "model_name", self.model_class.__name__.lower())
        self.verbose_name_plural = getattr(self.opts, "verbose_name_plural", inflector.plural(self.model_name).lower())
        self.ordering = getattr(self.opts, "ordering", None)
        self.unique_together = getattr(self.opts, "unique_together", ())

        self.app_label = getattr(self.opts, "app_label", None)
        if not self.app_label:
            app_config = apps.get_containing_app_config(self.model_class.__module__)
            self.app_label = getattr(app_config, "label", None) or "django_sorcery"

        self.label = "{}.{}".format(self.app_label, self.object_name)
        self.label_lower = "{}.{}".format(self.app_label, self.model_name)

        sa.event.listen(self.mapper, "mapper_configured", self._init)
        self._init(self.mapper, self.model_class)

    def _init(self, mapper, class_):
        assert mapper is self.mapper
        assert class_ is self.model_class

        self.primary_keys.clear()
        for col in self.mapper.primary_key:
            attr = self.mapper.get_property_by_column(col)
            self.primary_keys[attr.key] = column_info(col, attr, self, name=attr.key)

        self.properties.clear()
        for col in self.mapper.columns:
            attr = self.mapper.get_property_by_column(col)
            if attr.key not in self.primary_keys:
                self.properties[attr.key] = column_info(col, attr, self, name=attr.key)

        self.composites.clear()
        for composite in self.mapper.composites:
            self.composites[composite.key] = composite_info(getattr(self.model_class, composite.key), parent=self)
            for prop in composite.props:
                if prop.key in self.properties:
                    del self.properties[prop.key]

        self.relationships.clear()
        for relationship in self.mapper.relationships:
            self.relationships[relationship.key] = relation_info(relationship)

        self.local_fields = tuple(
            filter(
                lambda f: not f.name.startswith("_"),
                chain(self.primary_keys.values(), self.properties.values(), self.relationships.values()),
            )
        )
        self.fields = tuple(self.local_fields)
        self.concrete_fields = tuple(self.local_fields)
        self.column_properties = tuple(chain(self.primary_keys.items(), sorted(self.properties.items())))
        self.field_names = [
            attr
            for attr in chain(
                self.primary_keys.keys(), self.properties.keys(), self.composites.keys(), self.relationships.keys()
            )
            if not attr.startswith("_")
        ]

    def __dir__(self):
        return (
            dir(super())
            + list(vars(type(self)))
            + list(self.primary_keys)
            + list(self.properties)
            + list(self.composites)
            + list(self.relationships)
        )

    def __getattr__(self, name):
        if name in self.primary_keys:
            return self.primary_keys[name]
        if name in self.properties:
            return self.properties[name]
        if name in self.composites:
            return self.composites[name]
        if name in self.relationships:
            return self.relationships[name]
        return getattr(super(), name)

    def __repr__(self):
        reprs = ["<model_info({!s})>".format(self.model_class.__name__)]
        reprs.extend("    " + repr(i) for i in self.primary_keys.values())
        reprs.extend("    " + repr(i) for _, i in sorted(self.properties.items()))
        reprs.extend("    " + i for i in chain(*[repr(c).split("\n") for _, c in sorted(self.composites.items())]))
        reprs.extend("    " + repr(i) for _, i in sorted(self.relationships.items()))
        return "\n".join(reprs)

    @property
    def app_config(self):
        return apps.get_app_config(self.app_label) or apps.get_containing_app_config(self.model_name.__module__)

    def sa_state(self, instance):
        """Returns sqlalchemy instance state."""
        return sa.inspect(instance)

    def get_field(self, field_name):
        field = (
            self.primary_keys.get(field_name)
            or self.properties.get(field_name)
            or self.composites.get(field_name)
            or self.relationships.get(field_name)
        )
        if not field:
            raise FieldDoesNotExist

        return field

    def primary_keys_from_dict(self, kwargs):
        """Returns the primary key tuple from a dictionary to be used in a
        sqlalchemy query.get() call."""
        pks = []

        for attr, _ in self.primary_keys.items():
            pk = kwargs.get(attr)
            pks.append(pk)

        if any(pk is None for pk in pks):
            return None

        if len(pks) < 2:
            return next(iter(pks), None)

        return tuple(pks)

    def primary_keys_from_instance(self, instance):
        """Return a dict containing the primary keys of the ``instance``"""
        if instance is None:
            return None

        if len(self.primary_keys) > 1:
            return OrderedDict((name, getattr(instance, name)) for name in self.primary_keys)

        return getattr(instance, next(iter(self.primary_keys)))

    def get_key(self, instance):
        """Returns the primary key tuple from the ``instance``"""
        keys = self.mapper.primary_key_from_instance(instance)
        if any(key is None for key in keys):
            return
        return tuple(keys)

    def identity_key_from_instance(self, instance):
        """Returns the primary key tuple from the ``instance``"""
        keys = self.get_key(instance)
        if keys is None:
            return

        return Identity(self.model_class, self.get_key(instance))

    def identity_key_from_dict(self, kwargs):
        """Returns identity key from a dictionary for the given model."""
        pks = self.primary_keys_from_dict(kwargs)
        if pks is None:
            return

        return Identity(self.model_class, pks if isinstance(pks, tuple) else (pks,))

    def full_clean(self, instance, exclude=None, **kwargs):
        """Run model's full clean chain.

        This will run all of these in this order:

        * will validate all columns by using ``clean_<column>`` methods
        * will validate all nested objects (e.g. composites) with ``full_clean``
        * will run through all registered validators on ``validators`` attribute
        * will run full model validation with ``self.clean()``
        * if ``recursive`` kwarg is provided, will recursively clean all relations.
          Useful when all models need to be explicitly cleaned without flushing to DB.
        """
        exclude = exclude or []

        validators = [
            lambda x: getattr(x, "clean_fields", partial(self.clean_fields, instance=instance))(
                exclude=exclude, **kwargs
            ),
            lambda x: getattr(x, "clean_nested_fields", partial(self.clean_nested_fields, instance=instance))(
                exclude=exclude, **kwargs
            ),
            lambda x: getattr(x, "run_validators", partial(self.run_validators, instance=instance))(**kwargs),
            lambda x: x.clean(**kwargs),
        ]
        if kwargs.get("recursive", False):
            validators.append(
                lambda x: getattr(x, "clean_relation_fields", partial(self.clean_relation_fields, instance=instance))(
                    exclude=exclude, **kwargs
                )
            )

        runner = ValidationRunner(validators=validators)
        runner.is_valid(instance, raise_exception=True)

    def clean_fields(self, instance, exclude=None, **kwargs):
        """Clean all fields on object."""
        errors = {}
        exclude = exclude or []
        local_remote_pairs = set()
        for rel in self.relationships.values():
            for col in chain(*rel.local_remote_pairs):
                local_remote_pairs.add(col)

        props = [
            self.properties[prop]
            for prop in getattr(instance, "_get_properties_for_validation", lambda x: self.properties.values())()
            if prop in self.properties
        ]

        for f in props:
            raw_value = getattr(instance, f.name)
            is_blank = not bool(raw_value)
            is_nullable = f.null
            is_fk = f.column in local_remote_pairs
            is_defaulted = f.column.default or f.column.server_default
            is_required = f.required

            # skip validation if:
            # - field is blank and either when field is nullable so blank value is valid
            # - field has either local or server default value since we assume default value will pass validation
            #   since default values are assigned during flush which as after which validation is verified
            # - field is blank and is a foreign key in a relation that will be populated by the relation
            # - field is marked as not required in column info
            is_skippable = is_blank and (is_nullable or is_defaulted or is_fk or not is_required)

            if f.name in exclude or is_skippable:
                continue

            try:
                setattr(instance, f.name, f.clean(raw_value, instance))
            except ValidationError as e:
                errors[f.name] = e.error_list

        if errors:
            raise NestedValidationError(errors)

    def clean_nested_fields(self, instance, exclude=None, **kwargs):
        """Clean all nested fields which includes composites."""
        errors = {}
        exclude = exclude or []
        props = [
            self.composites[prop]
            for prop in getattr(instance, "_get_nested_objects_for_validation", lambda x: self.composites.values())()
            if prop in self.composites
        ]

        for f in props:
            try:
                e = exclude.get(f.name, [])
            except AttributeError:
                e = []

            # only exclude when subexclude is not either list or dict
            # otherwise validate nested object and let it ignore its own subfields
            is_nestable = e and isinstance(e, (dict, list, tuple))

            if f.name not in exclude or is_nestable:
                try:
                    f.full_clean(getattr(instance, f.name), exclude=e)
                except ValidationError as ex:
                    ex = NestedValidationError(ex)
                    errors[f.name] = ex.update_error_dict({})

        if errors:
            raise NestedValidationError(errors)

    def clean_relation_fields(self, instance, exclude=None, **kwargs):
        """Clean all relation fields."""
        exclude = exclude or []
        visited = kwargs.pop("visited", set())
        visited.add(id(instance))
        errors = {}

        props = getattr(instance, "_get_relation_objects_for_validation", lambda x: [])() or self.relationships

        for name in props:
            try:
                e = exclude.get(name, [])
            except AttributeError:
                e = []

            # only exclude when subexclude is not either list or dict
            # otherwise validate nested object and let it ignore its own subfields
            is_nestable = e and isinstance(e, (dict, list, tuple))

            # only validate relation if it is preloaded on the instance
            # otherwise full clean will explode to the complete model tree
            # which is not necessary as unloaded models cannot possibly
            # be in violating
            is_loaded = name in instance.__dict__

            if is_loaded and (name not in exclude or is_nestable):
                value = getattr(instance, name)

                if isinstance(value, (list, tuple)):
                    _errors = []

                    for i in value:
                        if id(i) not in visited:
                            try:
                                i.full_clean(exclude=e, visited=visited, **kwargs)
                            except AttributeError:
                                pass
                            except ValidationError as e:
                                e = NestedValidationError(e)
                                _errors.append(e.update_error_dict({}))

                    if _errors:
                        errors[name] = _errors

                else:
                    try:
                        if id(value) not in visited:
                            value.full_clean(exclude=e, visited=visited, **kwargs)
                    except AttributeError:
                        pass
                    except ValidationError as e:
                        e = NestedValidationError(e)
                        errors[name] = e.update_error_dict({})

        if errors:
            raise NestedValidationError(errors)

    def run_validators(self, instance, exclude=None, **kwargs):
        """Check all model validators registered on ``validators``
        attribute."""
        runner = ValidationRunner(validators=getattr(instance, "validators", []))
        runner.is_valid(instance, raise_exception=True)



"""Django REST Framework like model viewsets."""
from functools import update_wrapper
from inspect import getmembers

from django.core.exceptions import ImproperlyConfigured
from django.utils.decorators import classonlymethod
from django.views.generic.base import TemplateResponseMixin, View

from . import mixins


class GenericViewSet(TemplateResponseMixin, View):
    """Base class for all sqlalchemy model generic viewsets."""

    def get_template_names(self):
        self.template_name_suffix = f"_{self.action}"
        names = []

        try:
            names.extend(super().get_template_names())
        except ImproperlyConfigured:
            pass

        if hasattr(self, "get_model_template_name"):
            names.append(self.get_model_template_name())

        return names

    @classonlymethod
    def as_view(cls, actions=None, **initkwargs):
        # The suffix initkwarg is reserved for displaying the viewset type.
        # eg. 'List' or 'Instance'.
        cls.suffix = None

        # The detail initkwarg is reserved for introspecting the viewset type.
        cls.detail = None

        # Setting a basename allows a view to reverse its action urls. This
        # value is provided by the router through the initkwargs.
        cls.basename = None

        # actions must not be empty
        if not actions:
            raise TypeError(
                "The `actions` argument must be provided when "
                "calling `.as_view()` on a ViewSet. For example "
                "`.as_view({'get': 'list'})`"
            )

        # sanitize keyword arguments
        for key in initkwargs:
            if key in cls.http_method_names:
                raise TypeError(
                    "You tried to pass in the %s method name as a "
                    "keyword argument to %s(). Don't do that." % (key, cls.__name__)
                )

            if not hasattr(cls, key):
                raise TypeError("{}() received an invalid keyword {!r}".format(cls.__name__, key))

        def view(request, *args, **kwargs):
            self = cls(**initkwargs)
            # We also store the mapping of request methods to actions,
            # so that we can later set the action attribute.
            # eg. `self.action = 'list'` on an incoming GET request.
            self.action_map = actions
            method = request.method.lower()
            self.action = self.action_map.get(method, "metadata")

            # Bind methods to actions
            # This is the bit that's different to a standard view
            for method, action in actions.items():
                handler = getattr(self, action)
                setattr(self, method, handler)

            if hasattr(self, "get") and not hasattr(self, "head"):
                self.head = self.get

            self.request = request
            self.args = args
            self.kwargs = kwargs

            # And continue as usual
            return self.dispatch(request, *args, **kwargs)

        # take name and docstring from class
        update_wrapper(view, cls, updated=())

        # and possible attributes set by decorators
        # like csrf_exempt from dispatch
        update_wrapper(view, cls.dispatch, assigned=())

        # We need to set these on the view function, so that breadcrumb
        # generation can pick out these bits of information from a
        # resolved URL.
        view.cls = cls
        view.initkwargs = initkwargs
        view.suffix = initkwargs.get("suffix")
        view.actions = actions
        return view

    @classmethod
    def get_extra_actions(cls):
        """Get the methods that are marked as an extra ViewSet `@action`."""
        return [method for _, method in getmembers(cls, lambda attr: hasattr(attr, "bind_to_methods"))]


class ReadOnlyModelViewSet(mixins.ListModelMixin, mixins.RetrieveModelMixin, GenericViewSet):
    """A viewset that provides default `list()` and `retrieve()` actions.

    When used with router, it will map the following operations to actions on the viewset

    ====== ======================== =============== ======================
    Method Path                     Action          Route Name
    ====== ======================== =============== ======================
    GET    /                        list            <resource name>-list
    GET    /<pk>/                   retrieve        <resource name>-detail
    ====== ======================== =============== ======================
    """


class ModelViewSet(mixins.CreateModelMixin, mixins.UpdateModelMixin, mixins.DeleteModelMixin, ReadOnlyModelViewSet):
    """A viewset that provides default `new()`, `create()`, `retrieve()`,
    `edit()`, update()`, `confirm_destroy())`, `destroy()` and `list()`
    actions.

    When used with router, it will map the following operations to actions on the viewset

    ====== ======================== =============== ======================
    Method Path                     Action          Route Name
    ====== ======================== =============== ======================
    GET    /                        list            <resource name>-list
    POST   /                        create          <resource name>-list
    GET    /new/                    new             <resource name>-new
    GET    /<pk>/                   retrieve        <resource name>-detail
    POST   /<pk>/                   update          <resource name>-detail
    PUT    /<pk>/                   update          <resource name>-detail
    PATCH  /<pk>/                   update          <resource name>-detail
    DELETE /<pk>/                   destroy         <resource name>-detail
    GET    /<pk>/edit/              edit            <resource name>-edit
    GET    /<pk>/delete/            confirm_destoy  <resource name>-delete
    POST   /<pk>/delete/            destroy         <resource name>-delete
    ====== ======================== =============== ======================
    """



"""Current command."""
from functools import partial

import alembic

from ..alembic import AlembicCommand


class Current(AlembicCommand):
    """Shows current db revisions."""

    help = "Show current db revisions"

    def add_arguments(self, parser):
        parser.add_argument("app_label", nargs="?", help="App label of application to limit the output to.")

    def handle(self, app_label=None, verbosity=0, **kwargs):
        verbose = bool(verbosity - 1)
        appconfigs = [self.lookup_app(app_label)] if app_label is not None else self.sorcery_apps.values()

        for appconfig in appconfigs:
            self.stdout.write(self.style.SUCCESS(f"Revision for {appconfig.name} on database {appconfig.db.alias}"))
            with alembic.context.EnvironmentContext(
                appconfig.config,
                appconfig.script,
                fn=partial(self.display_version, verbose=verbose, appconfig=appconfig),
            ) as context:
                self.run_env(context, appconfig)

    def display_version(self, rev, context, verbose=False, appconfig=None):
        """Displays the alembic revision."""
        if verbose:
            self.stdout.write("Current revision(s) for {!r}".format(context.connection.engine.url))
        for revision in appconfig.script.get_all_current(rev):
            self.stdout.write(revision.cmd_format(verbose))

        return []


Command = Current



from contextlib import suppress

import sqlalchemy as sa
from django.db import DEFAULT_DB_ALIAS
from django.utils.module_loading import import_string

from .sqlalchemy import SQLAlchemy
from .transaction import TransactionContext
from .url import get_settings, make_url


class dbdict(dict):
    """Holds all configured :py:class:`..sqlalchemy.SQLAlchemy` instances."""

    def get(self, alias=None, cls=SQLAlchemy, **kwargs):
        """Returns a :py:class:`..sqlalchemy.SQLAlchemy` instance from
        configuration and registers it.

        Can return a custom
        :py:class:`..sqlalchemy.SQLAlchemy` instance thru args or thru ``SQLALCHEMY`` database setting in configuration.
        """
        alias = alias or DEFAULT_DB_ALIAS
        if alias in self:
            return self[alias]

        with suppress(Exception):
            settings = get_settings(alias)
            cls = import_string(settings.get("SQLALCHEMY"))

        assert SQLAlchemy in cls.mro(), "'%s' needs to subclass from SQLAlchemy" % cls.__name__

        url, _kwargs = make_url(alias)
        _kwargs.update(kwargs)
        _kwargs["alias"] = alias
        return self.setdefault(alias, cls(url, **_kwargs))

    def update(self, *args, **kwargs):
        for arg in args:
            other = dict(arg)
            for key in other:
                self[key] = other[key]
        for key in kwargs:
            self[key] = kwargs[key]

    def __setitem__(self, alias, val):
        if alias in self:
            raise RuntimeError("Database alias `{alias}` has already been created".format(alias=alias))

        if val in self.values():
            raise RuntimeError("Database alias `{alias}` has already been created".format(alias=alias))

        if not isinstance(val, SQLAlchemy):
            raise RuntimeError("Database alias `{alias}` has wrong type".format(alias=alias))

        super().__setitem__(alias, val)

    def rollback(self):
        """Applies rollback on all registered databases."""
        for db in self.values():
            db.rollback()

    def flush(self):
        """Applies flush on all registered databases."""
        for db in self.values():
            db.flush()

    def commit(self):
        """Applies commit on all registered databases."""
        for db in self.values():
            db.commit()

    def remove(self):
        """Applies remove on all registered databases."""
        for db in self.values():
            db.remove()

    def atomic(self, savepoint=True):
        """Returns a context manager/decorator that guarantee atomic execution
        of a given block or function across all configured and initialized
        SQLAlchemy instances."""
        return TransactionContext(*self.values(), savepoint=True)


def _index_foreign_keys(tbl):
    indexes = {tuple(sorted(col.name for col in ix.columns)) for ix in tbl.indexes}
    for fk in tbl.foreign_key_constraints:
        key = tuple(sorted(col.name for col in fk.columns))
        if key not in indexes:
            sa.Index(None, *list(fk.columns), use_alter=True)


def index_foreign_keys(*args):
    """Generates indexes for all foreign keys for a table or metadata
    tables."""
    for arg in args:
        if isinstance(arg, sa.Table):
            _index_foreign_keys(arg)
        elif isinstance(arg, sa.MetaData):
            for table in arg.tables.values():
                _index_foreign_keys(table)



from django_sorcery.db import databases


db = databases.get("test")


class OtherAppModel(db.Model):
    id = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String())


class OtherAppInOtherApp(db.Model):
    id = db.Column(db.Integer(), autoincrement=True, primary_key=True)
    name = db.Column(db.String())

    class Meta:
        app_label = "tests_testapp"


db.configure_mappers()
db.create_all()



"""
Signals
-------

Implements some basic signals using blinker
"""
from collections import defaultdict
from threading import local

import blinker


class ScopedSignal(blinker.NamedSignal):
    """Same as ``NamedSignal`` but signal is scoped to a thread.

    In other words, if a receiver is attached within a specific thread,
    even if signal is sent in another thread, in that other thread no
    receivers will be present and hence nothing will execute. Useful for
    adding one-off signal handlers for example to be executed at the end
    of unit-of-work (e.g. request) without adding a possibility that
    another thread might start executing the receiver.
    """

    def __init__(self, name, doc=None):
        self.name = name
        self.__doc__ = doc
        self.local = local()
        self.cleanup()

    @property
    def is_muted(self):
        return bool(self.local.__dict__.setdefault("is_muted", False))

    @is_muted.setter
    def is_muted(self, value):
        self.local.__dict__["is_muted"] = value  # pragma: nocover

    @property
    def receivers(self):
        """Return all thread scoped receivers."""
        return self.local.__dict__.setdefault("receivers", {})

    @property
    def _by_receiver(self):
        return self.local.__dict__.setdefault("_by_receiver", defaultdict(set))

    @property
    def _by_sender(self):
        return self.local.__dict__.setdefault("_by_sender", defaultdict(set))

    @property
    def _weak_senders(self):
        return self.local.__dict__.setdefault("_weak_senders", {})

    def cleanup(self):
        """Cleans up signal for the current thread scope."""
        self._clear_state()


class Namespace(blinker.Namespace):
    """A signal namespace that also manages scoped signals."""

    def scopedsignal(self, name, doc=None):
        """Returns the scoped signal for a given name."""
        try:
            return self[name]

        except KeyError:
            return self.setdefault(name, ScopedSignal(name, doc))

    @property
    def scoped_signals(self):
        """Returns all scoped signals."""
        for signal in self.values():
            if isinstance(signal, ScopedSignal):
                yield signal


all_signals = Namespace()

before_flush = all_signals.signal("before_flush")
after_flush = all_signals.signal("after_flush")

before_commit = all_signals.signal("before_commit")
before_scoped_commit = all_signals.scopedsignal("before_scoped_commit")

after_commit = all_signals.signal("after_commit")
after_scoped_commit = all_signals.scopedsignal("after_scoped_commit")

after_rollback = all_signals.signal("after_rollback")
after_scoped_rollback = all_signals.scopedsignal("after_scoped_rollback")

engine_created = all_signals.signal("engine_created")

before_middleware_request = all_signals.signal("before_middleware_request")
after_middleware_response = all_signals.signal("after_middleware_response")

declare_first = all_signals.signal("declare_first")
declare_last = all_signals.signal("declare_last")

before_create_all = all_signals.signal("before_create_all")
after_create_all = all_signals.signal("after_create_all")

before_drop_all = all_signals.signal("before_drop_all")
after_drop_all = all_signals.signal("after_drop_all")

alembic_app_created = all_signals.signal("alembic_app_created")
alembic_config_created = all_signals.signal("alembic_config_created")
alembic_include_object = all_signals.signal("alembic_include_object")
alembic_process_revision_directives = all_signals.signal("alembic_process_revision_directives")



from .__version__ import (
    __author__,
    __author_email__,
    __description__,
    __version__,
)


__all__ = [
    "__author__",
    "__author_email__",
    "__description__",
    "__version__",
]



import six

from django.test import TestCase
from django_sorcery.management.commands.sorcery_heads import Command

from .base import MIGRATION_DIR, MigrationMixin


M1 = '''"""zero

Revision ID: 000000000000
Revises:
Create Date: 2018-07-24 01:28:40.893136

"""


# revision identifiers, used by Alembic.
revision = '000000000000'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
'''

M2 = '''"""one

Revision ID: 000000000001
Revises: 000000000000
Create Date: 2018-07-24 02:02:55.504526

"""


# revision identifiers, used by Alembic.
revision = '000000000001'
down_revision = '000000000000'
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
'''


class TestHeads(MigrationMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.write_migration(M1, "{}_.py".format("000000000000"))
        self.write_migration(M2, "{}_.py".format("000000000001"))

    def tearDown(self):
        super().tearDown()
        for rev in ["000000000000", "000000000001"]:
            self.delete_migration("{}_.py".format(rev))

    def test(self):
        out = six.StringIO()
        cmd = Command(stdout=out)
        cmd.run_from_argv(["./manage.py sorcery", "heads", "--no-color"])

        out.seek(0)
        self.assertEqual(
            out.readlines(),
            [
                "Heads for tests_testapp on database test\n",
                "000000000001 (head) <tests_testapp> \n",
                "Heads for tests_otherapp on database test\n",
            ],
        )

    def test_verbose(self):
        out = six.StringIO()
        cmd = Command(stdout=out)
        cmd.run_from_argv(["./manage.py sorcery", "heads", "-v", "2", "--no-color"])

        out.seek(0)
        self.assertEqual(
            out.readlines(),
            [
                "Heads for tests_testapp on database test\n",
                "[tests_testapp]\n",
                "Rev: 000000000001 (head)\n",
                "Parent: 000000000000\n",
                "Path: {}/000000000001_.py\n".format(MIGRATION_DIR),
                "\n",
                "    one\n",
                "    \n",
                "    Revision ID: 000000000001\n",
                "    Revises: 000000000000\n",
                "    Create Date: 2018-07-24 02:02:55.504526\n",
                "Heads for tests_otherapp on database test\n",
            ],
        )



import os

from sqlalchemy.engine.url import make_url


def absjoin(*args):
    return os.path.normpath(os.path.abspath(os.path.join(*args)))


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

PROJECT_ROOT = os.path.dirname(absjoin(__file__))
STATIC_ROOT = absjoin(PROJECT_ROOT, "..", "static")

DB_URL = make_url(os.environ.get("DATABASE_URL", "postgresql://postgres:postgres@localhost"))

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "default_db",
        "USER": DB_URL.username,
        "PASSWORD": DB_URL.password,
        "HOST": DB_URL.host,
        "PORT": "",
    },
    "fromdbs": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "fromdbs",
        "USER": DB_URL.username,
        "PASSWORD": DB_URL.password,
        "HOST": DB_URL.host,
        "PORT": "",
    },
}

SQLALCHEMY_CONNECTIONS = {
    "default": {
        "DIALECT": "postgresql",
        "NAME": "default_db",
        "USER": DB_URL.username,
        "PASSWORD": DB_URL.password,
        "HOST": DB_URL.host,
        "PORT": "",
    },
    "test": {
        "DIALECT": "postgresql",
        "NAME": "test",
        "USER": DB_URL.username,
        "PASSWORD": DB_URL.password,
        "HOST": DB_URL.host,
        "PORT": "",
    },
    "minimal": {
        "DIALECT": "postgresql",
        "NAME": "minimal",
        "USER": DB_URL.username,
        "PASSWORD": DB_URL.password,
        "HOST": DB_URL.host,
        "PORT": "",
    },
    "minimal_backpop": {
        "DIALECT": "postgresql",
        "NAME": "minimal_backpop",
        "USER": DB_URL.username,
        "PASSWORD": DB_URL.password,
        "HOST": DB_URL.host,
        "PORT": "",
    },
    "terrible": {
        "DIALECT": "postgresql",
        "NAME": "minimal_backpop",
        "USER": DB_URL.username,
        "PASSWORD": DB_URL.password,
        "HOST": DB_URL.host,
        "PORT": "",
    },
}

INSTALLED_APPS = [
    "tests.minimalapp.apps.MinimalAppConfig",
    "tests.testapp.apps.TestAppConfig",
    "tests.otherapp.apps.OtherAppConfig",
    "django_sorcery",
    "django.contrib.staticfiles",
]

ROOT_URLCONF = "tests.urls"

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

USE_THOUSAND_SEPARATOR = True

SECRET_KEY = "secret"

ALLOWED_HOSTS = ["*"]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(os.path.dirname(__file__), "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
            ]
        },
    }
]



from contextlib import suppress

import sqlalchemy as sa
from django.conf import settings
from django_sorcery.db.query import Operation

from ..base import TestCase
from ..testapp.models import ModelOne, Owner, db


class TestSQLAlchemy(TestCase):
    def test_session(self):
        session = db.session()
        self.assertEqual(db.session(), db.session())
        self.assertFalse(session.autocommit)

        db.remove()
        session = db.session(autocommit=True)
        self.assertTrue(session.autocommit)

        with self.assertRaises(sa.exc.InvalidRequestError):
            db.session(autocommit=True)

    def test_callable(self):
        session = db()
        self.assertEqual(db(), db())
        self.assertFalse(session.autocommit)

        db.remove()
        session = db(autocommit=True)
        self.assertTrue(session.autocommit)

        with self.assertRaises(sa.exc.InvalidRequestError):
            db(autocommit=True)

    def test_url(self):
        self.assertEqual(db.bind.url, db.url)

    def test_repr(self):
        self.assertEqual(repr(db), f"<SQLAlchemy engine=postgresql://postgres:***@{settings.DB_URL.host}/test>")

    def test_queryproperty(self):
        qp = db.queryproperty(ModelOne)
        self.assertEqual(qp.model, ModelOne)
        self.assertEqual(qp.ops, [])

        qp = db.queryproperty(ModelOne, name="test")
        self.assertEqual(qp.model, ModelOne)
        self.assertEqual(qp.ops, [Operation(name="filter_by", args=(), kwargs={"name": "test"})])

    def test_inspector(self):
        inspector = db.inspector
        self.assertListEqual(inspector.get_schema_names(), ["information_schema", "public"])

    def test_make_args(self):
        self.assertEqual(db.args("test", other=True), ("test", {"other": True}))

    def test_atomic_decorator(self):
        @db.atomic()
        def do_something():
            db.add(Owner(first_name="test", last_name="last"))

        do_something()
        self.assertEqual(Owner.query.count(), 1)

    def test_atomic_decorator_exception(self):
        @db.atomic()
        def do_something():
            db.add(Owner(first_name="test", last_name="last"))
            raise Exception()

        with suppress(Exception):
            do_something()

        self.assertEqual(Owner.query.count(), 0)

    def test_atomic_context(self):
        with db.atomic():
            db.add(Owner(first_name="test", last_name="last"))

        self.assertEqual(Owner.query.count(), 1)

    def test_atomic_context_exception(self):
        with suppress(Exception), db.atomic():
            db.add(Owner(first_name="test", last_name="last"))
            raise Exception()

        self.assertEqual(Owner.query.count(), 0)



from bs4 import BeautifulSoup
from django import test
from django_sorcery.db import databases


try:
    from unittest import mock  # noqa
except ImportError:
    from unittest import mock  # noqa


class TestCase(test.TestCase):
    def setUp(self):
        super().setUp()
        self.maxDiff = None
        self.factory = test.RequestFactory()

    def tearDown(self):
        databases.rollback()
        databases.remove()
        super().tearDown()

    def get_soup(self, html):
        soup = BeautifulSoup(html, "html.parser")
        for select in soup.find_all("select"):
            # normalizing multiple attribute between django versions
            if select.get("multiple"):
                select["multiple"] = ""

        return soup



"""sqlalchemy query related things."""
from collections import namedtuple
from functools import partial

import sqlalchemy as sa
import sqlalchemy.orm  # noqa
from django.db.models.constants import LOOKUP_SEP

from ..utils import lower
from . import meta


Operation = namedtuple("Operation", ["name", "args", "kwargs"])

# todo add transforms support - e.g. column__date__gt
LOOKUP_TO_EXPRESSION = {
    "contains": lambda column, value: column.contains(value),
    # "date"
    # "day"
    "endswith": lambda column, value: column.endswith(value),
    "exact": lambda column, value: column == value,
    "gt": lambda column, value: column > value,
    "gte": lambda column, value: column >= value,
    # "hour"
    "icontains": lambda column, value: sa.func.lower(column).contains(lower(value)),
    "iendswith": lambda column, value: sa.func.lower(column).endswith(lower(value)),
    "iexact": lambda column, value: sa.func.lower(column) == lower(value),
    "iin": lambda column, value: sa.func.lower(column).in_(lower(i) for i in value),
    "in": lambda column, value: column.in_(value),
    # "iregex"
    "isnull": lambda column, value: column == None if value else column != None,  # noqa
    "istartswith": lambda column, value: sa.func.lower(column).startswith(lower(value)),
    "lt": lambda column, value: column < value,
    "lte": lambda column, value: column <= value,
    # "minute"
    # "month"
    # "quarter"
    "range": lambda column, value: column.between(*value),
    # "regex"
    # "second"
    "startswith": lambda column, value: column.startswith(value),
    # "time"
    # "week"
    # "week_day"
    # "year"
}


class Query(sa.orm.Query):
    """A customized sqlalchemy query."""

    def get(self, *args, **kwargs):
        """Return an instance based on the given primary key identifier, either
        as args or kwargs for composite keys.

        If no instance is found, returns ``None``.
        """
        if kwargs:
            mapper = self._only_full_mapper_zero("get")
            pk = meta.model_info(mapper).primary_keys_from_dict(kwargs)

            if pk is not None:
                return super().get(pk)

            return None

        return super().get(*args, **kwargs)

    def order_by(self, *criterion):
        """Standard SQLAlchemy ordering plus django-like expressions can be
        provided:

        For example::

            MyModel.objects.order_by("-id")
            MyModel.objects.order_by("name")
        """
        if all(isinstance(criteria, str) for criteria in criterion):
            model = self._only_full_mapper_zero("get")
            info = meta.model_info(model)
            new_criterion = []
            for criteria in criterion:
                direction = sa.asc
                if criteria[0] == "-":
                    criteria = criteria[1:]
                    direction = sa.desc
                elif criteria[0] == "+":
                    criteria = criteria[1:]

                col_info = info.primary_keys.get(criteria) or info.properties.get(criteria)
                new_criterion.append(direction(col_info.attribute))

            criterion = new_criterion

        return super().order_by(*criterion)

    def filter(self, *args, **kwargs):
        """Standard SQLAlchemy filtering plus django-like expressions can be
        provided:

        For example::

            MyModel.objects.filter(MyModel.id == 5)
            MyModel.objects.filter(id=5)
            MyModel.objects.filter(id__gte=5)
            MyModel.objects.filter(relation__id__gte=5)
        """
        args = args + tuple(self._lookup_to_expression(k, v) for k, v in kwargs.items())
        return super().filter(*args)

    def _lookup_to_expression(self, lookup, value):
        parts = lookup.split(LOOKUP_SEP)
        info = meta.model_info(self._only_full_mapper_zero("get"))

        props = dict(info.column_properties)
        lhs = None

        for i, part in enumerate(parts, 1):
            is_last = i == len(parts)

            if part in LOOKUP_TO_EXPRESSION:
                return LOOKUP_TO_EXPRESSION[part](lhs, value)

            elif part in info.relationships:
                rel = info.relationships[part]

                # directly comparing to model
                # e.g. .filter(relation=instance)
                if is_last:
                    return rel.attribute == value

                lhs = rel.related_model
                props = dict(meta.model_info(lhs).column_properties)

            elif part in info.composites:
                comp = info.composites[part]
                props = comp.properties

                # directly comparing to composite
                # e.g. .filter(composite=instance)
                if is_last:
                    return comp.attribute == value

            else:
                lhs = props[part].attribute

        return lhs == value


class QueryProperty:
    """A property class that returns a session scoped query object against the
    class when called. Used by the ``SQLAlchemy.queryproperty``

    For example::

        >>> class MyView(object):
        ...     queryset = db.queryproperty(FooModel)

    You can even pass default filtering criteria if needed::

        >>> class MyView(object):
        ...     queryset = db.queryproperty(FooModel, to_be_deleted=False)

    In addition this pattern can be used to implement Django's ORM style model managers::

        >>> class UserModel(db.Model):
        ...     id = db.Column(db.Integer(), primary_key=True)
        ...     username = db.Column(db.String())
        ...     is_active = db.Column(db.Boolean())
        ...
        ...     active = db.queryproperty(is_active=True)

    That can be used directly::

        >>> UserModel.metadata.create_all(bind=db.engine)
        >>> db.add_all([
        ...     UserModel(id=1, username='foo', is_active=False),
        ...     UserModel(id=2, username='bar', is_active=True),
        ... ])
        >>> db.flush()

        >>> UserModel.objects.all()
        [UserModel(id=1, is_active=False, username='foo'), UserModel(id=2, is_active=True, username='bar')]
        >>> UserModel.active.all()
        [UserModel(id=2, is_active=True, username='bar')]

    This pattern is very useful when combined with Django style views::

        >>> class MyView(object):
        ...     queryset = UserModel.active

        >>> MyView().queryset.all()
        [UserModel(id=2, is_active=True, username='bar')]

    Additional filters/options can be applied as well::

        >>> class MyView(object):
        ...     queryset = UserModel.active.filter(UserModel.username == 'test')

        >>> MyView().queryset.all()
        []
    """

    def __init__(self, db, model=None, *args, **kwargs):
        self.db = db
        self.model = model

        self.ops = []
        if args:
            self.ops.append(Operation("filter", args, {}))
        if kwargs:
            self.ops.append(Operation("filter_by", (), kwargs))

        # sanity checks
        if self.model:
            assert isinstance(self.model, type) and issubclass(
                self.model, self.db.Model
            ), "{!r} is not SQLAlchemy model subclassing {!r}".format(model, self.db.Model)

    def __repr__(self):
        return "<{} db={!r}, model={!r}>".format(self.__class__.__name__, self.db, self.model.__name__)

    def _with_op(self, name, *args, **kwargs):
        prop = type(self)(self.db, self.model)
        prop.ops += self.ops
        prop.ops.append(Operation(name, args, kwargs))
        return prop

    def __getattr__(self, item):
        if not hasattr(getattr(self.model, "query_class", Query), item):
            raise AttributeError("{!r} object has no attribute {!r}".format(self, item))

        return partial(self._with_op, item)

    def __get__(self, instance, owner):
        model = self.model or (owner if issubclass(owner, self.db.Model) else None)

        if not model:
            raise AttributeError(
                "Cannot access {} when not bound to a model. "
                "You can explicitly instantiate descriptor with model class - `db.queryproperty(Model)`."
                "".format(self.__class__.__name__)
            )

        try:
            mapper = sa.orm.class_mapper(model)
        except sa.orm.exc.UnmappedClassError:
            # when subclass references unmapped base class descriptor
            return

        query_class = getattr(model, "query_class", None) or self.db.query_class
        return self._apply_ops(query_class(mapper, session=self.db))

    def _apply_ops(self, query):
        for op in self.ops:
            query = getattr(query, op.name)(*op.args, **op.kwargs)
        return query



from django.urls import include, path
from django_sorcery.routers import SimpleRouter

from . import views


router = SimpleRouter()
router.register("", views.PollsViewSet)

app_name = "polls"
urlpatterns = [path("", include(router.urls))]



import sqlalchemy as sa
from django.core.exceptions import ImproperlyConfigured
from django_sorcery import fields
from django_sorcery.db import meta

from ...base import TestCase
from ...models_terrible_relations import Foo
from ...testapp.models import Owner, Part, Vehicle


class TestRelationshipMeta(TestCase):
    def test_relationship_meta(self):
        info = meta.model_info(Vehicle)

        rel = info.relationships["owner"]

        self.assertEqual(rel.related_model, Owner)
        self.assertEqual(rel.related_table, Owner.__table__)
        self.assertEqual(rel.parent_mapper, Vehicle.__mapper__)
        self.assertEqual(rel.related_mapper, Owner.__mapper__)
        self.assertEqual(rel.name, "owner")
        self.assertEqual(rel.direction, sa.orm.relationships.MANYTOONE)
        self.assertEqual(rel.foreign_keys, [Vehicle._owner_id.property.columns[0]])
        self.assertEqual(
            rel.local_remote_pairs, [(Vehicle._owner_id.property.columns[0], Owner.id.property.columns[0])]
        )
        self.assertEqual(
            rel.local_remote_pairs_for_identity_key,
            [(Vehicle._owner_id.property.columns[0], Owner.id.property.columns[0])],
        )
        self.assertFalse(rel.uselist)
        self.assertEqual(repr(rel), "<relation_info(Vehicle.owner)>")

    def test_relationship_meta_backref(self):
        info = meta.model_info(Owner)

        rel = info.relationships["vehicles"]

        self.assertEqual(rel.related_model, Vehicle)
        self.assertEqual(rel.related_table, Vehicle.__table__)
        self.assertEqual(rel.parent_mapper, Owner.__mapper__)
        self.assertEqual(rel.related_mapper, Vehicle.__mapper__)
        self.assertEqual(rel.name, "vehicles")
        self.assertEqual(rel.direction, sa.orm.relationships.ONETOMANY)
        self.assertEqual(rel.foreign_keys, [Vehicle._owner_id.property.columns[0]])
        self.assertEqual(
            rel.local_remote_pairs, [(Owner.id.property.columns[0], Vehicle._owner_id.property.columns[0])]
        )
        self.assertEqual(
            rel.local_remote_pairs_for_identity_key,
            [(Owner.id.property.columns[0], Vehicle._owner_id.property.columns[0])],
        )
        self.assertTrue(rel.uselist)
        self.assertEqual(repr(rel), "<relation_info(Owner.vehicles)>")

    def test_terrible_relationship_meta(self):
        info = meta.model_info(Foo)

        rel = info.relationships["partial_parent"]

        self.assertEqual(rel.related_model, Foo)
        self.assertEqual(rel.related_table, Foo.__table__)
        self.assertEqual(rel.name, "partial_parent")
        self.assertEqual(rel.direction, sa.orm.relationships.MANYTOONE)
        self.assertEqual(rel.foreign_keys, [Foo.parent_id2.property.columns[0]])
        self.assertEqual(rel.local_remote_pairs, [(Foo.parent_id2.property.columns[0], Foo.id2.property.columns[0])])
        self.assertEqual(
            rel.local_remote_pairs_for_identity_key,
            [
                (Foo.id1.property.columns[0], Foo.id1.property.columns[0]),
                (Foo.parent_id2.property.columns[0], Foo.id2.property.columns[0]),
            ],
        )
        self.assertFalse(rel.uselist)
        self.assertEqual(repr(rel), "<relation_info(Foo.partial_parent)>")

    def test_formfield(self):
        info = meta.model_info(Vehicle)

        with self.assertRaises(ImproperlyConfigured):
            info.owner.formfield()

        formfield = info.owner.formfield(session=Owner.query.session)
        self.assertIsInstance(formfield, fields.ModelChoiceField)

        formfield = info.parts.formfield(session=Part.query.session)
        self.assertIsInstance(formfield, fields.ModelMultipleChoiceField)

    def test_get_form_class(self):
        info = meta.model_info(Vehicle)

        self.assertEqual(info.parts.get_form_class(), fields.ModelMultipleChoiceField)
        self.assertEqual(info.owner.get_form_class(), fields.ModelChoiceField)

    def test_field_kwargs(self):
        info = meta.model_info(Vehicle)

        self.assertDictEqual(info.parts.field_kwargs, {"required": False})
        self.assertDictEqual(info.owner.field_kwargs, {"required": False})
